from typing import Dict, List

import numpy as np
import torch
from ppq.core import (CAFFE_DOMAIN, IS_DISPATCHED_GRAPH, NetworkFramework,
                      empty_ppq_cache)
from ppq.executor import TorchExecutor
from ppq.IR import (BaseGraph, GraphCommand, GraphCommandType, GraphFormatter,
                    GraphMerger, Operation, Opset, Variable)

from . import ppl_caffe_pb2

caffe_import_map = {}

from ppq.log import NaiveLogger

logger = NaiveLogger.get_logger('PPQ')


def build_temp_graph(initializer: Dict[str, dict], nodes: List[dict], inputs: List[str]=[], outputs: List[str]=[], graph: BaseGraph=None) -> BaseGraph:
    """could either build graph from scratch or append new nodes to a existing
    graph, if build form scratch, graph should be None, otherwise we use the
    given graph and append nodes and variables."""
    from_scratch = False
    if graph is None:
        graph = BaseGraph(name='Infer', built_from=NetworkFramework.CAFFE)
        from_scratch = True

    op_inputs_dict, op_outputs_dict = {}, {}
    for node in nodes:
        if node['name'] in graph.operations:
            raise KeyError(f"Duplicated operation {node['name']} was found.")
        graph.operations[node['name']] = Operation(name=node['name'], op_type=node['op_type'], 
                                    attributes=node['attribute'].copy(), opset=Opset(domain=CAFFE_DOMAIN, version=1))
        op_inputs_dict[node['name']] = [_ for _ in node['inputs']]
        op_outputs_dict[node['name']] = [_ for _ in node['outputs']]

    var_list = []
    for op_name, input_vars in op_inputs_dict.items():
        var_list.extend(input_vars)
    for op_name, output_vars in op_outputs_dict.items():
        var_list.extend(output_vars)

    # create all variable at once.
    for var_name in set(var_list):
        if var_name in graph.variables:
            continue
        graph.variables[var_name] = Variable(name=var_name)

    # build graph's input, output variables.
    # we only set graph inputs and outputs when build from scratch, i.e., when we are sure about graph inputs and outputs,
    # otherwise the graph inputs and outputs should manually set after the whole appending process from outside
    if from_scratch:
        try:
            for var_name in inputs:
                if var_name not in graph.variables: continue
                graph.inputs[var_name] = graph.variables[var_name]
            for var_name in outputs:
                graph.outputs[var_name] = graph.variables[var_name]
        except KeyError as e:
            raise KeyError(
                'seems you got an input/output variable that is not linked to any operation.')

    # build operation inputs, outputs variables.
    for op_name in op_inputs_dict:
        for var_name in op_inputs_dict[op_name]:
            var = graph.variables[var_name]
            var.dest_ops.append(graph.operations[op_name])
            graph.operations[op_name].inputs.append(graph.variables[var_name])
        for var_name in op_outputs_dict[op_name]:
            var = graph.variables[var_name]
            var.source_op = graph.operations[op_name]
            graph.operations[op_name].outputs.append(graph.variables[var_name])

    # initialize variable
    for var_name in initializer:
        if var_name in graph.variables:
            for dest_op in graph.variables[var_name].dest_ops:
                dest_op.parameters.append(graph.variables[var_name])
            graph.variables[var_name].value = initializer[var_name]['value']
            graph.variables[var_name].is_parameter = True

    return graph

def get_input_shape(net_def: ppl_caffe_pb2.NetParameter) -> Dict[str, list]:
    # Only support one format input shape, not support mixed format
    def layer_exist(layer_type):
        return layer_type in [item.type for item in net_def.layer]

    input_shape = {k: None for k in net_def.input}
    # Given input shape use input_shape field
    if len(net_def.input_shape) != 0:
        for i, name in enumerate(net_def.input):
            input_shape[name] = list(net_def.input_shape[i].dim)
    # Given input shape use input_dim
    # TODO: Here only support  4-D input
    elif len(net_def.input_dim) != 0:
        for i, name in enumerate(net_def.input):
            input_shape[name] = list(net_def.input_dim[i * 4:(i + 1) * 4])
    # Given input shape use input layer
    elif layer_exist('Input'):
        input_layer = [item for item in net_def.layer]
        for layer in input_layer:
            input_shape[layer.top[0]] = list(layer.input_param.shape.dim)
    else:
        raise TypeError('Unsupported network input format.')

    for k, v in input_shape.items():
        if v is None:
            raise TypeError("shape of input '%s' is not specified." % k)

    return input_shape

def register_class(cls):
    caffe_import_map[cls.__name__] = cls
    return cls


class CaffeOpBuilder(object):
    caffe_to_onnx = {
        'Convolution': 'Conv',
        'BatchNorm': 'BatchNormalization',
        'BN': 'BatchNormalization',
        'Deconvolution': 'ConvTranspose',
        'ReLU6': 'Clip',
        'PReLU': 'PRelu',
        'InnerProduct': 'Gemm',
        'ReflectionPad': 'Pad',
        'NNUpsample': 'Resize',
        'TanH': 'Tanh',
        'Power': 'Pow',
        'InstanceNorm': 'InstanceNormalization',
        'SubpixelUp': 'DepthToSpace',
        'SubpixelDown': 'SpaceToDepth',
        'ArgMax': 'CaffeArgMax',
        'HSwish': 'HardSwish',
        'HSigmoid': 'HardSigmoid',
        'MatMul': 'MatMul'
    }

    def __init__(self,
                graph: ppl_caffe_pb2.NetParameter,
                layer: ppl_caffe_pb2.LayerParameter,
                input_shape: List[list]
    ):
        self.graph = graph
        self.layer = layer
        self.settings = {}
        self.initializer = {}
        self.nodes = []
        self.op_type = self.get_type()

        assert len(self.layer.bottom) == len(input_shape), 'Given input_shape and layer.bottom have different length'
        self.input_shape = input_shape
        self.out_shape = None
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

    def get_type(self):
        op_type = self.layer.type
        if self.layer.type in CaffeOpBuilder.caffe_to_onnx:
            op_type = CaffeOpBuilder.caffe_to_onnx[self.layer.type]

        return op_type

    def get_attr(self):
        return {}

    def trans(self):
        sub_int, param_name = self.param_process()
        self.initializer.update(sub_int)
        self.nodes.append({'name': self.layer.name, 'inputs': list(self.layer.bottom) + param_name,
                           'outputs': list(self.layer.top), 'op_type': self.op_type, 'attribute': self.get_attr()})
        self.infer_shape()
        return build_temp_graph(self.initializer, self.nodes, graph=self.graph)

    @empty_ppq_cache
    def infer_shape(self):
        input_names = self.layer.bottom
        # if len(set(self.layer.bottom)) != len(self.layer.bottom):
        #     # There are same variable in bottom, need to de_inplace
        #     visited_var = {}
        #     for var in self.layer.bottom:
        #         if var in visited_var.keys():
        #             input_names.append(var + '_' + str(visited_var[var]))
        #             visited_var[var] += 1
        #         else:
        #             input_names.append(var)
        #             visited_var[var] = 1
        # else:
        #     input_names = self.layer.bottom
        input_value = [np.random.rand(*i).astype('f') for i in self.input_shape]
        input_value = [torch.from_numpy(tensor).to(self.device) for tensor in input_value]
        input_value = dict([(x, y) for x, y in zip(input_names, input_value)])
        temp_graph = build_temp_graph(self.initializer, self.nodes, input_names, list(self.layer.top))

        # BUG FIX 12162021, Format graph when loading caffe model.
        formatter = GraphFormatter(GraphMerger(temp_graph))
        formatter(GraphCommand(GraphCommandType.FORMAT_CONSTANT_INPUT))
        formatter(GraphCommand(GraphCommandType.FUSE_BN))
        formatter(GraphCommand(GraphCommandType.FORMAT_PARAMETERS))
        formatter(GraphCommand(GraphCommandType.FORMAT_CAST))
        formatter(GraphCommand(GraphCommandType.DELETE_ISOLATED))

        # PATCH 20220805, add dispatching tag for executing.
        temp_graph.set_extension_attrib(IS_DISPATCHED_GRAPH, True)
        executor = TorchExecutor(graph=temp_graph, device=self.device)
        outputs = executor.forward(input_value, list(self.layer.top))
        self.out_shape = [list(i.shape) for i in outputs]

    def param_process(self):
        initializer = {}
        param_name = []
        for i, tensor in enumerate(self.layer.blobs):
            init_name = self.layer.name + '_param_' + str(i)
            dims = tensor.shape.dim
            if self.layer.type == 'BN' and len(tensor.shape.dim) == 4:
                # Some BN layers give the dim in [1, channel, 1, 1], while some give the dim in [channel]
                dims = tensor.shape.dim[1:2]
            np_type = 'float32' if not isinstance(tensor.data[0], int) else 'int'
            value = np.array(tensor.data, dtype=np_type).reshape(dims)
            initializer[init_name] = {'dims': dims, 'value': value}
            param_name.append(init_name)
        return initializer, param_name


@register_class
class Convolution(CaffeOpBuilder):
    def get_attr(self):
        conv = self.layer.convolution_param
        if conv.HasField('kernel_h') and conv.HasField('kernel_w'):
            kernel_shape = [conv.kernel_h, conv.kernel_w]
        else:
            kernel_shape = [conv.kernel_size] * 2
        if conv.HasField('stride_h') and conv.HasField('stride_w'):
            strides = [conv.stride_h, conv.stride_w]
        else:
            strides = [conv.stride] * 2
        if conv.HasField('pad_h') and conv.HasField('pad_w'):
            pads = [conv.pad_h, conv.pad_w] * 2
        else:
            pads = [conv.pad] * 4
        self.settings['kernel_shape'] = kernel_shape
        self.settings['strides'] = strides
        self.settings['pads'] = pads
        group = conv.group
        # convolution layer should not use hole parameter
        if conv.hole_h != 1 or conv.hole_w != 1:
            logger.warning(f'{self.layer.name} is dilated convolution, check if the real backend supports')

        self.settings['dilations'] = [conv.hole_h, conv.hole_w]
        self.settings['group'] = group
        return self.settings


@register_class
class BatchNorm(CaffeOpBuilder):
    # caffe batchnorm layer, only compute x_norm = (x-u)/std
    # Thus set weights = 1 and bias = 0
    def get_attr(self):
        batchnorm = self.layer.batch_norm_param
        # If use_global_stats is not given, the CaffeOpBuilder value of protobuf is false,
        # but PPL change it to be True in the code space
        if not batchnorm.use_global_stats:
            logger.warning('Now cannot convert BatchNorm layer with use_global_stats==False. Change it to True')
        # Because only support use_global_stats == True, we do not parse 'moving_average_fraction' here
        # self.settings['momentum'] = batchnorm.moving_average_fraction
        self.settings['epsilon'] = batchnorm.eps
        return self.settings

    def trans(self):
        # Set weights = 1
        dim = self.layer.blobs[0].shape.dim
        weights = ppl_caffe_pb2.BlobProto()
        weights.data.extend(np.ones(dim, dtype='f'))
        weights.shape.dim.extend(dim)
        self.layer.blobs.insert(0, weights)

        # Set bias = 0
        bias = ppl_caffe_pb2.BlobProto()
        bias.data.extend(np.zeros(dim, dtype='f'))
        bias.shape.dim.extend(dim)
        self.layer.blobs.insert(1, bias)

        if len(self.layer.blobs) == 5:
            # Remove scale_factor
            scale_factor = np.array(self.layer.blobs[4].data, dtype='f')
            self.layer.blobs.remove(self.layer.blobs[4])

            # Update running_mean and running_var according to scale_factor
            running_mean = self.layer.blobs[2]
            new_mean = np.array(running_mean.data, dtype='f') / scale_factor
            running_mean.ClearField('data')
            running_mean.data.extend(new_mean)

            running_var = self.layer.blobs[3]
            new_var = np.array(running_var.data, dtype='f') / scale_factor
            running_var.ClearField('data')
            running_var.data.extend(new_var)

        return super(BatchNorm, self).trans()


@register_class
class BN(CaffeOpBuilder):
    # BatchNorm + Scale, the same with BatchNormalization in pytorch
    # y = alpha * ((x-u)/std) + beta
    def get_attr(self):
        batchnorm = self.layer.bn_param
        if batchnorm.moving_average:
            logger.warning('Now cannot convert BatchNorm layer with moving_average==True. Change it to False')
        self.settings['epsilon'] = batchnorm.var_eps
        # Because only support moving_average == False, we do not parse 'decay' in BN layer
        return self.settings


@register_class
class Deconvolution(Convolution):
    pass


@register_class
class ReLU(CaffeOpBuilder):
    def get_type(self):
        if self.layer.relu_param.HasField('negative_slope'):
            return 'LeakyRelu'
        else:
            return 'Relu'

    def get_attr(self):
        relu = self.layer.relu_param
        if relu.HasField('negative_slope'):
            self.op_type = 'LeakyRelu'
            self.settings['alpha'] = relu.negative_slope
        return self.settings


@register_class
class PReLU(CaffeOpBuilder):
    def trans(self):
        assert len(self.layer.blobs) == 1, 'PReLU lacks of slope'
        dim_info = self.layer.blobs[0].shape.dim
        # squeeze the PReLU slope [1, 128, 1, 1] to [1, 128] if input is 2-D
        # prelu only has one input data
        if len(dim_info) == 4 and len(self.input_shape[0]) == 2:
            dim_info.pop()
            dim_info.pop()
        return super(PReLU, self).trans()


@register_class
class Concat(CaffeOpBuilder):
    def get_attr(self):
        concat = self.layer.concat_param
        self.settings['axis'] = concat.axis
        return self.settings


@register_class
class Softmax(CaffeOpBuilder):
    def get_attr(self):
        softmax = self.layer.softmax_param
        self.settings['axis'] = softmax.axis
        return self.settings


@register_class
class Transpose(CaffeOpBuilder):
    def get_attr(self):
        transpose = self.layer.transpose_param
        self.settings['perm'] = transpose.dim[:]
        return self.settings


@register_class
class ReduceL2(CaffeOpBuilder):
    def get_attr(self):
        reducel2 = self.layer.reducel2_param
        self.settings['axes'] = reducel2.axes
        self.settings['keepdims'] = reducel2.keepdims
        return self.settings


@register_class
class Reduce(CaffeOpBuilder):
    def get_type(self):
        return 'ReduceMean'

    def get_attr(self):
        reduce = self.layer.reduce_param
        self.settings['axis'] = reduce.axis
        return self.settings


@register_class
class Div(CaffeOpBuilder):
    pass


@register_class
class Pooling(CaffeOpBuilder):
    def get_attr(self):
        pool = self.layer.pooling_param
        if pool.global_pooling:
            return self.settings

        if pool.HasField('kernel_h') and pool.HasField('kernel_w'):
            kernel_shape = [pool.kernel_h, pool.kernel_w]
        else:
            kernel_shape = [pool.kernel_size] * 2
        if pool.HasField('stride_h') and pool.HasField('stride_w'):
            strides = [pool.stride_h, pool.stride_w]
        else:
            strides = [pool.stride] * 2
        if pool.HasField('pad_h') and pool.HasField('pad_w'):
            pads = [pool.pad_h, pool.pad_w] * 2
        else:
            pads = [pool.pad] * 4
        self.settings['kernel_shape'] = kernel_shape
        self.settings['strides'] = strides
        self.settings['pads'] = pads
        self.settings['ceil_mode'] = int(pool.ceil_mode)
        if not pool.global_pooling and pool.pool == 1:
            # set count_include_pad for average pooling
            self.settings['count_include_pad'] = 1
        return self.settings

    def get_type(self):
        # MAX=0, AVE=1
        pool = self.layer.pooling_param
        if not pool.global_pooling:
            if pool.pool == 0:
                return 'MaxPool'
            elif pool.pool == 1:
                return 'AveragePool'
            else:
                raise TypeError('unknown pooling operation type')

        else:
            if pool.pool == 0:
                return 'GlobalMaxPool'
            elif pool.pool == 1:
                return 'GlobalAveragePool'
            else:
                raise TypeError('unknown pooling operation type')


@register_class
class Eltwise(CaffeOpBuilder):
    def get_type(self):
        # PROD/MUL=0, ADD=1, MAX=2
        param = self.layer.eltwise_param
        if param.operation == 0:
            return 'Mul'
        elif param.operation == 1:
            return 'Add'
        elif param.operation == 2:
            return 'Max'
        else:
            raise TypeError('unknown eltwise operation type')

    def trans(self):
        def _add_mul(_node_name, input_name, factor):
            new_node_name = _node_name + '_mul'
            out_name = new_node_name + '_out'
            factor_name = new_node_name + '_coeff'
            init = {factor_name: {'dims': [1], 'value': np.array(factor, dtype='f')}}
            node = {'name': new_node_name, 'inputs': [input_name, factor_name], 'outputs': [out_name],
                    'op_type': 'Mul', 'attribute': self.get_attr()}
            self.initializer.update(init)
            self.nodes.append(node)
            return out_name

        coeff = list(self.layer.eltwise_param.coeff)
        if len(self.layer.bottom) > len(self.layer.eltwise_param.coeff):
            coeff.extend([1.0] * (len(self.layer.bottom) - len(self.layer.eltwise_param.coeff)))
        elif len(self.layer.bottom) < len(self.layer.eltwise_param.coeff):
            logger.error(f'Less bottom than coeff in {self.layer.name}')
            exit(-1)
        bottom_var = list(self.layer.bottom)

        # The number of inputs is uncertain, thus here may generate a series of binary op
        last = bottom_var[0]
        if coeff[0] != 1.0:
            node_name = self.layer.name + '_0'
            output_name = _add_mul(node_name, last, coeff[0])
            last = output_name

        for i, name in enumerate(bottom_var[1:]):
            node_name = self.layer.name + '_' + str(i + 1)
            output_name = node_name + '_out' if (i + 2 != len(bottom_var)) else self.layer.top[0]
            op_type = self.op_type
            if coeff[i + 1] == -1.0:
                op_type = 'Sub'
            elif coeff[i + 1] != 1.0:
                mul_op_name = _add_mul(node_name, name, coeff[i + 1])
                name = mul_op_name
            sub_node = {'name': node_name, 'inputs': [last, name], 'outputs': [output_name], 'op_type': op_type,
                        'attribute': self.get_attr()}
            last = output_name
            self.nodes.append(sub_node)

        if self.nodes[-1]['outputs'] != self.layer.top:
            # If Eltwise op only has one bottom and coeff != 1.0, the output name will be changed and
            # lead to missing key error
            assert len(self.nodes[-1]['outputs']) == len(self.layer.top)
            self.nodes[-1]['outputs'] = list(self.layer.top)
            logger.debug(f'Output of {self.nodes[-1]} != {self.layer.top}, Change it!')

        self.infer_shape()
        return build_temp_graph(self.initializer, self.nodes, graph=self.graph)


@register_class
class Reshape(CaffeOpBuilder):
    def trans(self):
        dim = self.layer.reshape_param.shape.dim
        init_name = self.layer.name + '_param'
        self.initializer = {init_name: {'dims': [len(dim)], 'value': np.array(dim, dtype='int64')}}
        self.nodes = [{'name': self.layer.name, 'inputs': list(self.layer.bottom) + [init_name],
                       'outputs': list(self.layer.top), 'op_type': self.layer.type, 'attribute': {}}]
        self.infer_shape()
        return build_temp_graph(self.initializer, self.nodes, graph=self.graph)


@register_class
class ReLU6(CaffeOpBuilder):
    def get_attr(self):
        self.settings = {'min': 0.0, 'max': 6.0}
        return self.settings

@register_class
class Clip(CaffeOpBuilder):
    def get_attr(self):
        param = self.layer.clip_param
        self.settings = {'min': param.min, 'max': param.max}
        return self.settings

@register_class
class Mul(CaffeOpBuilder):
    pass

@register_class
class Add(CaffeOpBuilder):
    pass

@register_class
class InnerProduct(CaffeOpBuilder):
    def get_attr(self):
        self.settings = {'transB': 1}
        if self.layer.inner_product_param.HasField('axis'):
            self.settings['axis'] = self.layer.inner_product_param.axis
        else:
            self.settings['axis'] = 1
        return self.settings

    def trans(self):
        ip_param = self.layer.inner_product_param
        assert ip_param.num_output == self.layer.blobs[0].shape.dim[0], 'Given num_output is not same with weight shape'
        input_var_list = list(self.layer.bottom)

        '''
        if not (len(self.input_shape[0]) == 2 and ip_param.axis == 1):
            # Insert an reshape op whether needed or not due to no shape information
            flatten_shape = [0, -1]
            reshape_name = self.layer.name + '_flatten_param'
            output_name = self.layer.name + '_flatten_output'
            self.initializer[reshape_name] = {'dims': [2],
                                              'value': np.array(flatten_shape, dtype='int64')}
            self.nodes.append({'name': self.layer.name + '_flatten', 'inputs': input_var_list + [reshape_name],
                               'outputs': [output_name], 'op_type': 'Reshape', 'attribute': {}})
            input_var_list = [output_name]
        '''

        # process the shape of bias, let [num_output, 1] to [num_output]
        if len(self.layer.blobs) == 2:
            bias_shape = self.layer.blobs[1].shape.dim
            if bias_shape == [ip_param.num_output, 1]:
                self.layer.blobs[1].shape.dim.pop()

        sub_int, param_name = self.param_process()
        self.initializer.update(sub_int)
        self.nodes.append({'name': self.layer.name, 'inputs': input_var_list + param_name,
                           'outputs': list(self.layer.top), 'op_type': self.op_type, 'attribute': self.get_attr()})
        self.infer_shape()
        return build_temp_graph(self.initializer, self.nodes, graph=self.graph)


@register_class
class ReflectionPad(CaffeOpBuilder):
    def get_attr(self):
        pad = self.layer.pad_param
        self.settings = {'pads': [pad.pad_h, pad.pad_w] * 2, 'mode': 'reflect'}
        return self.settings


@register_class
class NNUpsample(CaffeOpBuilder):
    # Upsample is deprecated in onnx-11, it's covered by Resize
    def trans(self):
        # Only support input is 4-D
        param = self.layer.nn_upsample_param
        resize = param.resize
        roi = self.layer.name + '_roi'
        scales = self.layer.name + '_scales'
        self.initializer = {roi: {'dims': [4], 'value': np.array([0, 0, 0, 0], dtype='f')},
                            scales: {'dims': [4], 'value': np.array([1, 1, resize, resize], dtype='f')}}

        self.nodes = [{'name': self.layer.name, 'inputs': list(self.layer.bottom) + [roi, scales],
                       'outputs': list(self.layer.top), 'op_type': self.op_type, 'attribute': {'mode': 'nearest'}}]

        self.infer_shape()
        return build_temp_graph(self.initializer, self.nodes, graph=self.graph)


@register_class
class Sigmoid(CaffeOpBuilder):
    pass


@register_class
class Slice(CaffeOpBuilder):
    def trans(self):
        param = self.layer.slice_param
        axis = param.axis
        axis_name = self.layer.name + '_axes'
        self.initializer[axis_name] = {'dims': [1], 'value': np.array([axis], dtype='int64')}

        slice_points = [0] + list(param.slice_point) + [self.input_shape[0][axis]]
        splits = [(slice_points[i], slice_points[i + 1]) for i in range(len(slice_points) - 1)]
        for i, pairs in enumerate(splits):
            starts = self.layer.name + '_starts_' + str(i)
            ends = self.layer.name + '_ends_' + str(i)
            self.initializer[starts] = {'dims': [1], 'value': np.array([pairs[0]], dtype='int64')}
            self.initializer[ends] = {'dims': [1], 'value': np.array([pairs[1]], dtype='int64')}
            self.nodes.append(
                {'name': self.layer.name + '_' + str(i), 'inputs': list(self.layer.bottom) + [starts, ends, axis_name],
                 'outputs': [list(self.layer.top)[i]], 'op_type': self.op_type, 'attribute': {}})
        self.infer_shape()
        return build_temp_graph(self.initializer, self.nodes, graph=self.graph)


@register_class
class Flatten(CaffeOpBuilder):
    def get_attr(self):
        flatten = self.layer.flatten_param
        self.settings['axis'] = flatten.axis
        if flatten.end_axis not in [-1, len(self.input_shape) - 1]:
            logger.error('Only support flattening inputs into 2-D matrix')
            exit(-1)
        return self.settings


@register_class
class Interp(CaffeOpBuilder):
    def trans(self):
        self.initializer = {}
        self.nodes = [{'name': self.layer.name, 'inputs': list(self.layer.bottom),
                       'outputs': list(self.layer.top), 'op_type': self.op_type, 'attribute': self.get_attr()}]
        self.infer_shape()
        return build_temp_graph(self.initializer, self.nodes, graph=self.graph)

    def get_attr(self):
        interp = self.layer.interp_param
        self.settings['shrink_factor'] = interp.shrink_factor
        self.settings['zoom_factor'] = interp.zoom_factor
        self.settings['pad_beg'] = interp.pad_beg
        self.settings['pad_end'] = interp.pad_end
        self.settings['height'] = interp.height
        self.settings['width'] = interp.width
        self.settings['align_corners'] = interp.align_corners
        self.settings['mode'] = 'linear'
        # For caffe ops which don't mapping to onnx compute in torch_backend.py
        # input_shape will be saved in attributes for dump_to_onnx func
        self.settings['input_shape'] = self.input_shape
        return self.settings


@register_class
class Tile(CaffeOpBuilder):
    def get_attr(self):
        tile = self.layer.tile_param
        self.settings['axis'] = tile.axis
        self.settings['tiles'] = tile.tiles
        return self.settings


@register_class
class SubpixelDown(CaffeOpBuilder):
    def get_attr(self):
        subpixel_down = self.layer.subpixel_down_param
        self.settings['blocksize'] = subpixel_down.downsample
        return self.settings


@register_class
class SubpixelUp(CaffeOpBuilder):
    def get_attr(self):
        subpixel_up = self.layer.subpixel_up_param
        self.settings['blocksize'] = subpixel_up.upsample
        return self.settings


@register_class
class Scale(CaffeOpBuilder):
    def get_attr(self):
        scale_param = self.layer.scale_param
        self.settings['axis'] = scale_param.axis
        self.settings['num_axes'] = scale_param.num_axes
        self.settings['bias_term'] = scale_param.bias_term
        # For caffe ops which don't mapping to onnx compute in torch_backend.py
        # input_shape will be saved in attributes for dump_to_onnx func
        self.settings['input_shape'] = self.input_shape
        return self.settings


@register_class
class TanH(CaffeOpBuilder):
    # TODO
    # parameter 'engine' is  not added yet
    pass


@register_class
class Power(CaffeOpBuilder):
    def get_attr(self):
        power_param = self.layer.power_param
        self.settings['power'] = power_param.power
        self.settings['scale'] = power_param.scale
        self.settings['shift'] = power_param.shift
        return self.settings


@register_class
# !Crop in ppl_caffe is different from caffe
class Crop(CaffeOpBuilder):
    def get_attr(self):
        crop_param = self.layer.crop_param
        self.settings['crop_param'] = crop_param
        return self.settings


@register_class
class ChannelShuffle(CaffeOpBuilder):
    def get_attr(self):
        channel_shuffle_param = self.layer.channel_shuffle_param
        self.settings['group'] = channel_shuffle_param.group
        # For caffe ops which don't mapping to onnx compute in torch_backend.py
        # input_shape will be saved in attributes for dump_to_onnx func
        self.settings['input_shape'] = self.input_shape
        return self.settings


@register_class
class InstanceNorm(CaffeOpBuilder):
    def get_attr(self):
        instance_norm_param = self.layer.instance_norm_param
        self.settings['num_features'] = instance_norm_param.num_features
        self.settings['eps'] = instance_norm_param.eps
        self.settings['affine'] = instance_norm_param.affine
        return self.settings


@register_class
class Parameter(CaffeOpBuilder):
    def get_attr(self):
        parameter_param = self.layer.parameter_param
        self.settings['m'] = parameter_param.m
        self.settings['n'] = parameter_param.n
        self.settings['batch'] = parameter_param.batch
        self.settings['channel'] = parameter_param.channel
        self.settings['height'] = parameter_param.height
        self.settings['width'] = parameter_param.width
        return self.settings


@register_class
class ArgMax(CaffeOpBuilder):
    def get_attr(self):
        argmax_param = self.layer.argmax_param
        self.settings['out_max_val'] = argmax_param.out_max_val
        self.settings['top_k'] = argmax_param.top_k
        self.settings['axis'] = argmax_param.axis if argmax_param.HasField('axis') else None
        return self.settings

@register_class
class HSwish(CaffeOpBuilder):
    pass

@register_class
class HSigmoid(CaffeOpBuilder):
    pass

@register_class
class MatMul(CaffeOpBuilder):
    pass




import numpy as np
import torch

import ppq.lib as PFL
from ppq import TargetPlatform, TorchExecutor, graphwise_error_analyse
from ppq.api import ENABLE_CUDA_KERNEL, export_ppq_graph, load_onnx_graph
from ppq.quantization.optim import *

calibration_dataloader = [torch.rand([1, 3, 640, 640]) for _ in range(32)]

with ENABLE_CUDA_KERNEL():
    graph = load_onnx_graph('Models/yolov5s.v5.onnx')

    quantizer   = PFL.Quantizer(platform=TargetPlatform.OPENVINO_INT8, graph=graph) # 取得 OPENVINO_INT8 所对应的量化器
    dispatching = PFL.Dispatcher(graph=graph).dispatch(                             # 生成调度表
        quant_types=quantizer.quant_operation_types)

    # ------------------------------------------------------------
    # Yolo5 前面可能有一坨 Slice, Concat 算子
    # 后面可能有一坨后处理算子，我们不希望量化它们，你可用下面的方法将它们解除量化
    # 不同模型中层的名字可能不一样，你需要按照你的模型对它们进行手动修改
    # ------------------------------------------------------------

    # Concat_40 往前的所有算子不量化
    from ppq.IR import SearchableGraph
    search_engine = SearchableGraph(graph)
    for op in search_engine.opset_matching(
        sp_expr=lambda x: x.name == 'Concat_40',
        rp_expr=lambda x, y: True,
        ep_expr=None, direction='up'
    ):
        dispatching[op.name] = TargetPlatform.FP32

    # Sigmoid_280 往后的所有算子不量化
    # Sigmoid_246 往后的所有算子不量化
    # Sigmoid_314 往后的所有算子不量化
    for op in search_engine.opset_matching(
        sp_expr=lambda x: x.name in {'Sigmoid_246', 'Sigmoid_280', 'Sigmoid_314'},
        rp_expr=lambda x, y: True,
        ep_expr=None, direction='down'
    ):
        dispatching[op.name] = TargetPlatform.FP32

    # 为算子初始化量化信息
    for op in graph.operations.values():
        quantizer.quantize_operation(
            op_name = op.name, platform = dispatching[op.name])

    # 初始化执行器
    collate_fn = lambda x: x.to('cuda')
    executor = TorchExecutor(graph=graph, device='cuda')
    executor.tracing_operation_meta(inputs=torch.zeros(size=[1, 3, 640, 640]).cuda())
    executor.load_graph(graph=graph)

    # ------------------------------------------------------------
    # 创建优化管线，由于后续还要继续训练我们的模型，我们不能在此处调用
    # ParameterBakingPass()，一旦模型权重完成烘焙，则它们不能被进一步调整
    # ------------------------------------------------------------
    pipeline = PFL.Pipeline([
        QuantizeSimplifyPass(),
        QuantizeFusionPass(
            activation_type=quantizer.activation_fusion_types),
        ParameterQuantizePass(),
        RuntimeCalibrationPass(),
        PassiveParameterQuantizePass(),
        QuantAlignmentPass(force_overlap=True),
        ParameterBakingPass()
    ])

with ENABLE_CUDA_KERNEL():
    # 调用管线完成量化
    pipeline.optimize(
        graph=graph, dataloader=calibration_dataloader, verbose=True, 
        calib_steps=32, collate_fn=collate_fn, executor=executor)

    graphwise_error_analyse(
        graph=graph, running_device='cuda', dataloader=calibration_dataloader, 
        collate_fn=lambda x: x.cuda())

    export_ppq_graph(
        graph=graph, platform=TargetPlatform.TRT_INT8, 
        graph_save_to='Output/quantized.onnx', 
        config_save_to='Output/quantized.json')

from ppq.utils.TensorRTUtil import Benchmark, Profiling, build_engine

build_engine(
    onnx_file='Output/quantized.onnx', 
    engine_file='Output/quantized.engine', 
    int8=True, int8_scale_file='Output/quantized.json')
Benchmark('Output/quantized.engine')
Profiling('Output/quantized.engine')


from ppq import *
from ppq.IR.morph import GraphDecomposer
from ppq.api import *
import torch

graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Gemm', name='gemm', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(), graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, 10]))],
                       outputs=[graph.create_variable()])
graph.create_operation(op_type='Softmax', name='softmax', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, ]))], 
                       outputs=[graph.create_variable()])
processor = GraphDecomposer(graph)
processor.decompose_gemm()

assert len(graph.operations) == 2
assert len(graph.operations['gemm'].inputs) == 2
assert graph.operations['gemm'].type == 'Matmul'

graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Gemm', name='gemm', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(), 
                               graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, 10])),
                               graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, 10]))],
                       outputs=[graph.create_variable()])
graph.create_operation(op_type='Softmax', name='softmax', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, ]))], 
                       outputs=[graph.create_variable()])
processor = GraphDecomposer(graph)
processor.decompose_gemm()

assert len(graph.operations) == 3
assert len(graph.operations['gemm'].inputs) == 2
assert graph.operations['gemm'].type == 'Matmul'


graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Gemm', name='gemm', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(),
                               graph.create_variable(is_parameter=True, value=torch.ones(size=[10, 10])),
                               graph.create_variable(is_parameter=True, value=torch.ones(size=[10, 10]))],
                       attributes={'alpha': 2, 'beta': 3},
                       outputs=[graph.create_variable()])
graph.create_operation(op_type='Softmax', name='softmax', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, ]))], 
                       outputs=[graph.create_variable()])
processor = GraphDecomposer(graph)
processor.decompose_gemm()

assert len(graph.operations) == 3
assert len(graph.operations['gemm'].inputs) == 2
assert graph.operations['gemm'].type == 'Matmul'
assert graph.operations['gemm'].inputs[1].value.mean().item() == 2


graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Gemm', name='gemm', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(),
                               graph.create_variable(is_parameter=True, value=torch.ones(size=[10, 10])),
                               graph.create_variable(is_parameter=True, value=torch.ones(size=[10, 10]))],
                       attributes={'transA': 0, 'transB': 1},
                       outputs=[graph.create_variable()])
graph.create_operation(op_type='Softmax', name='softmax', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, ]))], 
                       outputs=[graph.create_variable()])
processor = GraphDecomposer(graph)
processor.decompose_gemm()

assert len(graph.operations) == 3
assert len(graph.operations['gemm'].inputs) == 2
assert graph.operations['gemm'].type == 'Matmul'

try:
    graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
    matmul = \
    graph.create_operation(op_type='Gemm', name='gemm', 
                        platform=TargetPlatform.UNSPECIFIED, 
                        inputs=[graph.create_variable(),
                                graph.create_variable(is_parameter=True, value=torch.ones(size=[10, 10])),
                                graph.create_variable(is_parameter=True, value=torch.ones(size=[10, 10]))],
                        attributes={'transA': 1, 'transB': 0},
                        outputs=[graph.create_variable()])
    graph.create_operation(op_type='Softmax', name='softmax', platform=TargetPlatform.UNSPECIFIED, 
                        inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, ]))], 
                        outputs=[graph.create_variable()])
    processor = GraphDecomposer(graph)
    processor.decompose_gemm()
except ValueError as e:
    pass


from ppq import BaseGraph, TorchExecutor
import torch
graph = BaseGraph(name='TestBed')

# Create a Gemm
gemm = graph.create_operation(
    op_type='Gemm', 
    inputs=[graph.create_variable()], 
    outputs=[graph.create_variable()]
)

graph.create_variable(
    value=torch.rand(size=[32, 32]), 
    is_parameter=True,
    dest_ops=[gemm])

add = graph.create_operation(
    op_type='Add', 
    inputs=[gemm.outputs[0]], 
    outputs=[graph.create_variable()])

graph.create_variable(
    value=torch.rand(size=[1, 32]), 
    is_parameter=True, 
    dest_ops=[add])

graph.mark_variable_as_graph_input(gemm.inputs[0])
graph.mark_variable_as_graph_output(add.outputs[0])

sample = torch.rand(size=[8, 32]).cuda()
executor = TorchExecutor(graph=graph)
ref_out = executor.forward(inputs=sample)[0]

from ppq.IR import GraphMerger
GraphMerger(graph).fuse_bias_add()
executor = TorchExecutor(graph=graph)
out = executor.forward(inputs=sample)[0]

for op in graph.operations.values():
    assert op.type != 'Add', 'Bias Add Fusion Failed.'
assert torch.max(ref_out - out).item() < 1e-7, 'Bias Add Fusion Failed.'

# Create a Gemm
graph = BaseGraph(name='TestBed')
gemm = graph.create_operation(
    op_type='Gemm', 
    inputs=[graph.create_variable()], 
    outputs=[graph.create_variable()]
)

graph.create_variable(
    value=torch.rand(size=[32, 32]), 
    is_parameter=True,
    dest_ops=[gemm])

add = graph.create_operation(
    op_type='Add', 
    inputs=[gemm.outputs[0]], 
    outputs=[graph.create_variable()])

graph.create_variable(
    value=torch.rand(size=[8, 32]), 
    is_parameter=True, 
    dest_ops=[add])

graph.mark_variable_as_graph_input(gemm.inputs[0])
graph.mark_variable_as_graph_output(add.outputs[0])

sample = torch.rand(size=[8, 32]).cuda()
executor = TorchExecutor(graph=graph)
ref_out = executor.forward(inputs=sample)[0]

from ppq.IR import GraphMerger
GraphMerger(graph).fuse_bias_add()
executor = TorchExecutor(graph=graph)
out = executor.forward(inputs=sample)[0]

for op in graph.operations.values():
    assert op.type != 'Add', 'Bias Add Fusion Failed.'
assert torch.max(ref_out - out).item() < 1e-7, 'Bias Add Fusion Failed.'

# Create a Conv
graph = BaseGraph(name='TestBed')
conv = graph.create_operation(
    op_type='Conv', attributes={
        'dilations': [1, 1], 
        'group': 1, 
        'kernel_shape': [3, 3], 
        'pads': [1, 1, 1, 1], 
        'strides': [1, 1]
    },
    inputs=[graph.create_variable()], 
    outputs=[graph.create_variable()]
)

graph.create_variable(
    value=torch.rand(size=[16, 3, 3, 3]), 
    is_parameter=True,
    dest_ops=[conv])

add = graph.create_operation(
    op_type='Add', 
    inputs=[conv.outputs[0]], 
    outputs=[graph.create_variable()])

graph.create_variable(
    value=torch.rand(size=[1, 16, 1, 1]), 
    is_parameter=True, 
    dest_ops=[add])

graph.mark_variable_as_graph_input(conv.inputs[0])
graph.mark_variable_as_graph_output(add.outputs[0])

sample = torch.rand(size=[1, 3, 32, 32]).cuda()
executor = TorchExecutor(graph=graph)
ref_out = executor.forward(inputs=sample)[0]

from ppq.IR import GraphMerger
GraphMerger(graph).fuse_bias_add()
executor = TorchExecutor(graph=graph)
out = executor.forward(inputs=sample)[0]

for op in graph.operations.values():
    assert op.type != 'Add', 'Bias Add Fusion Failed.'
assert torch.max(ref_out - out).item() < 1e-7, 'Bias Add Fusion Failed.'


from typing import Union

import torch
from ppq.core import (PASSIVE_OPERATIONS, OperationQuantizationConfig,
                      QuantizationPolicy, QuantizationProperty,
                      QuantizationStates, RoundingPolicy, TargetPlatform)
from ppq.IR import BaseGraph, Operation

from .base import BaseQuantizer


class GraphCoreQuantizer(BaseQuantizer):
    def __init__(
        self, graph: BaseGraph
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits    = 8
        self._exponent_bits  = 4
        self._quant_min      = - 448.0
        self._quant_max      = + 448.0

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy, rounding=self.rounding_policy,
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=self._exponent_bits,
            quant_max=self._quant_max, quant_min=self._quant_min, observer_algorithm='floating')

        # 一些特殊的算子需要更复杂的量化逻辑，我们在下面的代码中对量化信息进行调整
        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul'}:

            # 将 Conv 的参数设置为对称 per-channel 量化
            if operation.type in {'Conv', 'ConvTranspose'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.PER_CHANNEL + 
                    QuantizationProperty.FLOATING +
                    QuantizationProperty.POWER_OF_2
                )
                # 量化的 channel_axis 对于 conv 而言是 0，对于反卷积而言是 1
                # 参数量化使用 min max observe
                conv_weight_config.channel_axis = (1 if operation.type == 'ConvTranspose' else 0)
                conv_weight_config.observer_algorithm = 'floating'

            # 将 Gemm 的参数设置为对称 per-channel 量化
            elif operation.type in {'Gemm', 'MatMul'}:
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.PER_CHANNEL +
                    QuantizationProperty.FLOATING +
                    QuantizationProperty.POWER_OF_2
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'floating'

            # 如果有 bias, 那 bias 就不量化了，因为我也不知道应该怎么量化他
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # 有一些算子是被动量化的，他们的输入输出将共享量化 scale
            base_quant_config.is_active_quant_op = False

        return base_quant_config

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.PPL_CUDA_INT8

    @ property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @ property
    def quant_operation_types(self) -> set:
        return {
            'Conv', 'Relu', 'PRelu', 'Clip', 'Gemm',
            'Resize', 'MaxPool', 'AveragePool',
            'GlobalMaxPool', 'GlobalAveragePool',
            'Mul', 'Add', 'LeakyRelu', 'Split', 'Concat',
            'Transpose', 'Slice', 'Reshape', 'Flatten',
            'Sigmoid', 'ReduceMean', 'ConvTranspose',
            'MatMul'
        }

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL +
            QuantizationProperty.PER_TENSOR +
            QuantizationProperty.FLOATING +
            QuantizationProperty.POWER_OF_2
        )

    @ property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @ property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Sigmoid', 'LeakyRelu', 'Swish', 'Mish', 'PRelu'}



class TensorRTQuantizer_FP8(BaseQuantizer):
    def __init__(
        self, graph: BaseGraph
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits    = 8
        self._exponent_bits = 4
        self._quant_min      = - 448.0
        self._quant_max      = + 448.0

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy, rounding=self.rounding_policy,
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=self._exponent_bits,
            quant_max=self._quant_max, quant_min=self._quant_min, observer_algorithm='floating')

        # 一些特殊的算子需要更复杂的量化逻辑，我们在下面的代码中对量化信息进行调整
        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul'}:

            # 将 Conv 的参数设置为对称 per-channel 量化
            if operation.type in {'Conv', 'ConvTranspose'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.PER_CHANNEL + 
                    QuantizationProperty.FLOATING +
                    QuantizationProperty.POWER_OF_2
                )
                # 量化的 channel_axis 对于 conv 而言是 0，对于反卷积而言是 1
                # 参数量化使用 min max observe
                conv_weight_config.channel_axis = (1 if operation.type == 'ConvTranspose' else 0)
                conv_weight_config.observer_algorithm = 'floating'

            # 将 Gemm 的参数设置为对称 per-channel 量化
            elif operation.type in {'Gemm', 'MatMul'}:
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.PER_CHANNEL +
                    QuantizationProperty.FLOATING +
                    QuantizationProperty.POWER_OF_2
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'floating'

            # 如果有 bias, 那 bias 就不量化了，因为我也不知道应该怎么量化他
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32
                
            # 所有算子只量化输入
            for output_config in base_quant_config.output_quantization_config:
                output_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # 有一些算子是被动量化的，他们的输入输出将共享量化 scale
            base_quant_config.is_active_quant_op = False

        return base_quant_config

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.PPL_CUDA_INT8

    @ property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @ property
    def quant_operation_types(self) -> set:
        return {
            'Conv', 'Gemm',
            'AveragePool',
            'GlobalAveragePool',
            'ConvTranspose',
            'MatMul'
        }

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL +
            QuantizationProperty.PER_TENSOR +
            QuantizationProperty.FLOATING +
            QuantizationProperty.POWER_OF_2
        )

    @ property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @ property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Sigmoid', 'LeakyRelu', 'Swish', 'Mish', 'PRelu'}



import os
import sys
from typing import List

import numpy as np
import torch

from ppq.core import (
    DataType,
    OperationQuantizationConfig,
    QuantizationProperty,
    QuantizationStates,
    QuantizationVisibility,
    TargetPlatform,
    TensorQuantizationConfig,
    convert_any_to_numpy,
    convert_any_to_torch_tensor,
)
from ppq.executor.base import OPERATION_FORWARD_TABLE
from ppq.IR import BaseGraph, Operation, OperationExporter, Variable
from ppq.IR.quantize import QuantableOperation
from ppq.log import NaiveLogger
from ppq.parser.espdl.espdl_typedef import (
    ACTIVATION_OP_SET,
    MATH_OP_SET,
    QUANT_EXCLUDE_OP_SET,
    QUANT_OP_SET,
    EspQuantType,
    ExporterPatternInfo,
    LayoutAnnotation,
)
from ppq.quantization.qfunction.linear import PPQLinearQuant_toInt
from ppq.utils.round import ppq_tensor_round

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), ".")))

logger = NaiveLogger.get_logger("ESPDL")


class EspdlQuantHelper:
    """Helper class for processing onnx qdq format"""

    @staticmethod
    def TQC_Exportable_Check(
        TQC: TensorQuantizationConfig, bounded_var: Variable
    ) -> bool:
        if not TQC.can_export(True):
            logger.warning(
                f"Warning: skip {bounded_var.name} because it's not exportable"
            )
            return False

        if TQC.visibility == QuantizationVisibility.INTERNAL:
            logger.warning(
                f"Warning: skip {bounded_var.name} because TAC visibility is internal"
            )
            return False

        if TQC.num_of_bits == 8 and TQC.policy.has_property(
            QuantizationProperty.LINEAR
        ):
            if TQC.policy.has_property(QuantizationProperty.ASYMMETRICAL):
                range_check = TQC.quant_max <= 255 and TQC.quant_min >= 0
            else:
                range_check = TQC.quant_max <= 127 and TQC.quant_min >= -128
        else:
            range_check = True

        if not range_check:
            logger.warning(
                f"Is it not safe to export TQC({bounded_var.name}) to Onnx, "
                f"INT8 value range must be [-128, 127] or [0, 255], "
                f"however [{TQC.quant_min, TQC.quant_max}] was given."
            )
            return False
        return True


def fuse_downstream_operation(
    graph: BaseGraph,
    fusing_downstream_op: Operation,
    keep_coherence: bool = False,
    remove_unlinked_variable: bool = False,
):
    """Remove operation from graph, this function will unlink removing
    operation from current graph, pop it from graph.operations, and remove
    it from all its input and output variables.

    Parameters of this removing operations will be removed from graph by this function, without warning.

    Args:
        fusing_downstream_op (Operation): [description]

        keep_coherence (bool): if keep_coherence = True,
            PPQ will link downstream operations of removing op to the upstream operation.
            if there is more than 1 input and output variable, ppq will link input[0] with output[0]
    """
    if fusing_downstream_op.name not in graph.operations:
        raise KeyError(
            f"Can not remove operation {fusing_downstream_op.name}, operation not found."
        )

    # removing all parameters first.
    for parameter in fusing_downstream_op.inputs.copy():
        if keep_coherence and fusing_downstream_op.type in {"Constant", "Identity"}:
            break
        if parameter.is_parameter:
            parameter.dest_ops.clear()
            parameter.value = None  # clear memory.
            fusing_downstream_op.inputs.remove(parameter)

            graph.variables.pop(parameter.name)

    related_vars = [
        var for var in fusing_downstream_op.inputs + fusing_downstream_op.outputs
    ]
    input_var, output_var = (
        fusing_downstream_op.inputs[0]
        if fusing_downstream_op.num_of_input >= 1
        else None,
        fusing_downstream_op.outputs[0]
        if fusing_downstream_op.num_of_output >= 1
        else None,
    )

    # remove operation from its output variables
    for _output_var in fusing_downstream_op.outputs:
        _output_var.source_op = None
    fusing_downstream_op.outputs.clear()

    # remove operation from its input variables
    for _input_var in fusing_downstream_op.inputs:
        if fusing_downstream_op in _input_var.dest_ops:
            _input_var.dest_ops.remove(fusing_downstream_op)
    fusing_downstream_op.inputs.clear()

    if input_var is not None and output_var is not None and keep_coherence:
        removing_var = input_var
        source_op = removing_var.source_op
        source_op.outputs[source_op.outputs.index(removing_var)] = output_var
        output_var.source_op = source_op
        removing_var.source_op = None
        removing_var.dest_ops.clear()
        graph.remove_variable(removing_var)

    graph.operations.pop(fusing_downstream_op.name)

    if remove_unlinked_variable:
        for var in related_vars:
            if (
                var.source_op is None
                and len(var.dest_ops) == 0
                and var.name in graph.variables
            ):
                graph.remove_variable(var)

    return graph


def infer_qtype(config: TensorQuantizationConfig):
    offset_dtype, value_dtype = torch.int8, torch.int8
    if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
        offset_dtype = torch.uint8
        value_dtype = torch.uint8
    if config.num_of_bits > 8:
        offset_dtype = torch.int16
        value_dtype = torch.int16
    elif config.num_of_bits > 16:
        offset_dtype = torch.int32
        value_dtype = torch.int32
    return offset_dtype, value_dtype


class InsertQuantTypePattern(OperationExporter):
    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if (
            op.platform == TargetPlatform.ESPDL_INT8
            or op.platform == TargetPlatform.ESPDL_S3_INT8
        ):
            op.attributes["quant_type"] = EspQuantType.S8
        elif (
            op.platform == TargetPlatform.ESPDL_INT16
            or op.platform == TargetPlatform.ESPDL_S3_INT16
        ):
            op.attributes["quant_type"] = EspQuantType.S16
        else:
            op.attributes["quant_type"] = EspQuantType.F32

        return op


class InsertQuantNodePattern(OperationExporter):
    @staticmethod
    def insert_quantize_node(
        graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation
    ) -> Operation:
        """
        Insert a Quantize Node on given variable, according to given TensorQuantizationConfig.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> INT
            offset_dtype, value_type = infer_qtype(config)
            scale = convert_any_to_torch_tensor(
                config.scale.clone(), dtype=torch.float32
            )
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type="QuantizeLinear", attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes["axis"] = config.channel_axis
            else:
                created.attributes["axis"] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(
                    f"Unexpected Error in Exporting Op {op.name}({op.type})."
                )

            graph.create_variable(
                name=None, value=scale, is_parameter=True, dest_ops=[created]
            )
            graph.create_variable(
                name=None, value=offset, is_parameter=True, dest_ops=[created]
            )

            created.outputs[0].dtype = value_type
            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created

        else:
            raise TypeError(
                f"PPQ Can not export quantization information with variable {var.name}, "
                "Unexpected Quantization property."
            )

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in QUANT_OP_SET or not isinstance(op, QuantableOperation):
            return op

        for config, var in [_ for _ in op.config_with_variable]:
            inserting_op, inserting_var = op, var
            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not var.is_parameter:
                if var.source_op:
                    if var.source_op.type in QUANT_OP_SET:
                        assert (
                            var.source_op.num_of_input == 3
                        ), "Quantize Node Format Error, need as least 3 inputs."
                        assert isinstance(var.source_op, Operation)
                        continue
                    elif var in op.inputs:
                        if (
                            not isinstance(var.source_op, QuantableOperation)
                            and var.source_op.type not in QUANT_EXCLUDE_OP_SET
                        ):
                            logger.debug(
                                f"Insert Quantize Node for {op.name}:{var.name}"
                            )
                            InsertQuantNodePattern.insert_quantize_node(
                                graph=graph,
                                var=inserting_var,
                                config=config,
                                op=inserting_op,
                            )
        return op


class InsertRequantNodePattern(OperationExporter):
    @staticmethod
    def insert_requantize_node(
        graph: BaseGraph,
        var: Variable,
        upstream_config: TensorQuantizationConfig,
        config: TensorQuantizationConfig,
        op: Operation,
    ) -> Operation:
        """
        Insert a ReQuantize Node on given variable, according to given TensorQuantizationConfig.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            upstream_offset_dtype, upstream_value_type = infer_qtype(upstream_config)
            upstream_scale = convert_any_to_torch_tensor(
                upstream_config.scale.clone(), dtype=torch.float32
            )
            upstream_offset = ppq_tensor_round(upstream_config.offset.clone()).type(
                torch.float
            )
            offset_dtype, value_type = infer_qtype(config)
            scale = convert_any_to_torch_tensor(
                config.scale.clone(), dtype=torch.float32
            )
            offset = ppq_tensor_round(config.offset.clone()).type(torch.float)

            created = graph.create_operation(op_type="RequantizeLinear", attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes["axis"] = config.channel_axis
            else:
                created.attributes["axis"] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(
                    f"Unexpected Error in Exporting Op {op.name}({op.type})."
                )

            rescale = scale / upstream_scale
            reoffset = ppq_tensor_round(
                offset - ppq_tensor_round(upstream_offset / rescale, config.rounding)
            ).type(offset_dtype)

            graph.create_variable(
                name=None, value=rescale, is_parameter=True, dest_ops=[created]
            )
            graph.create_variable(
                name=None, value=reoffset, is_parameter=True, dest_ops=[created]
            )

            created.inputs[0].dtype = upstream_value_type
            created.inputs[0].shape = var.shape
            created.outputs[0].shape = var.shape
            created.outputs[0].dtype = value_type
            return created

        else:
            raise TypeError(
                f"PPQ Can not export quantization information with variable {var.name}, "
                "Unexpected Quantization property."
            )

    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in QUANT_OP_SET or not isinstance(op, QuantableOperation):
            return op

        for config, var in [_ for _ in op.config_with_variable]:
            inserting_op, inserting_var = op, var
            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not var.is_parameter:
                if var.source_op:
                    if var.source_op.type in QUANT_OP_SET:
                        assert (
                            var.source_op.num_of_input == 3
                        ), "Quantize Node Format Error, need as least 3 inputs."
                        assert isinstance(var.source_op, Operation)
                        continue
                    elif var in op.inputs and isinstance(
                        var.source_op, QuantableOperation
                    ):
                        source_op_output_var_index = var.source_op.outputs.index(var)
                        source_op_output_config = var.source_op.output_quant_config[
                            source_op_output_var_index
                        ]
                        scale_diff = torch.max(
                            torch.abs(source_op_output_config.scale - config.scale)
                        ).item()
                        zeropoint_diff = torch.max(
                            torch.abs(source_op_output_config.offset - config.offset)
                        ).item()

                        if (
                            source_op_output_config.num_of_bits != config.num_of_bits
                            or scale_diff >= 1e-4
                            or zeropoint_diff >= 1e-1
                        ):
                            # if config
                            logger.debug(
                                f"Insert Requantize Node for {op.name}:{var.name}"
                            )
                            InsertRequantNodePattern.insert_requantize_node(
                                graph=graph,
                                var=inserting_var,
                                upstream_config=source_op_output_config,
                                config=config,
                                op=inserting_op,
                            )

        return op


class InsertDequantNodePattern(OperationExporter):
    @staticmethod
    def insert_dequantize_node(
        graph: BaseGraph, var: Variable, config: TensorQuantizationConfig, op: Operation
    ) -> Operation:
        """
        Insert a DeQuantize Node on given variable, according to given TensorQuantizationConfig.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            offset_dtype, value_type = infer_qtype(config)
            scale = convert_any_to_torch_tensor(
                config.scale.clone(), dtype=torch.float32
            )
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type="DequantizeLinear", attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes["axis"] = config.channel_axis
            else:
                created.attributes["axis"] = None

            if var in op.inputs:
                graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(
                    f"Unexpected Error in Exporting Op {op.name}({op.type})."
                )

            graph.create_variable(
                name=None, value=scale, is_parameter=True, dest_ops=[created]
            )
            graph.create_variable(
                name=None, value=offset, is_parameter=True, dest_ops=[created]
            )

            created.inputs[0].dtype = value_type
            created.inputs[0].shape = var.shape
            created.outputs[0].shape = var.shape
            created.outputs[0].dtype = torch.float32
            return created

        else:
            raise TypeError(
                f"PPQ Can not export quantization information with variable {var.name}, "
                "Unexpected Quantization property."
            )

    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in QUANT_OP_SET or not isinstance(op, QuantableOperation):
            return op

        for config, var in [_ for _ in op.config_with_variable]:
            inserting_op, inserting_var = op, var
            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not var.is_parameter:
                if var in op.outputs:
                    for dest_op in var.dest_ops:
                        if (
                            dest_op
                            and dest_op.type not in QUANT_OP_SET
                            and not isinstance(dest_op, QuantableOperation)
                            and dest_op.type not in QUANT_EXCLUDE_OP_SET
                        ):
                            logger.debug(
                                f"Insert Dequantize Node for {op.name}:{var.name}"
                            )
                            InsertDequantNodePattern.insert_dequantize_node(
                                graph=graph,
                                var=inserting_var,
                                config=config,
                                op=dest_op,
                            )
        return op


class InsertPreNodeOfMatMulPattern(OperationExporter):
    @staticmethod
    def insert_transpose_node(
        graph: BaseGraph, var: Variable, op: Operation, perm: List[int]
    ) -> Operation:
        """
        Insert a Transpose Node on given variable, according to given perm.
        """
        created = None
        if op and perm != range(len(perm)):
            logger.debug(
                f"insert transpose node: op: {op.name}, var:{var.name}, perm:{perm}"
            )
            if var in op.inputs:
                created = graph.create_operation(op_type="Transpose", attributes={"perm": perm})
                var_index = op.inputs.index(var)
                if isinstance(op, QuantableOperation):
                    # For transpose op,  input_quantization_config == output_quantization_config
                    new_config = OperationQuantizationConfig(
                        [op.input_quant_config[var_index]],
                        [op.input_quant_config[var_index]],
                    )
                    created = QuantableOperation(created, new_config, op.platform)
                    graph.operations[created.name] = created

                graph.insert_op_before(A=created, B=op, input_idx=var_index)
                new_var = created.outputs[0]
                new_var.shape = [var.shape[i] for i in perm]
                new_var.is_parameter = False
                new_var.dtype = var.dtype

            else:
                raise ValueError(f"Unexpected Error in Exporting Op {op.name}({op.type}).")

        return created


    @staticmethod
    def insert_reshape_node(
        graph: BaseGraph, 
        var: Variable,
        op: Operation, 
        shape: List[int], 
        allowzero: int = 0
    ) -> Operation:
        """
        Insert a Reshape Node on given variable, according to given TensorQuantizationConfig.
        """
        created = None
        if op and var in op.inputs:
            logger.debug(
                f"insert reshape node: op: {op.name}, var: {var.name}, shape: {shape}"
            )
            created = graph.create_operation(op_type="Reshape", attributes={"allowzero": allowzero})

            var_index = op.inputs.index(var)
            if isinstance(op, QuantableOperation):
                # For reshape op,  input_quantization_config[0] == output_quantization_config
                shape_config = op.input_quant_config[var_index].copy()
                shape_config.state = QuantizationStates.FP32
                shape_config.observer_algorithm = "percentile"

                new_config = OperationQuantizationConfig(
                    [op.input_quant_config[var_index], shape_config],
                    [op.input_quant_config[var_index]],
                )
                created = QuantableOperation(created, new_config, op.platform)
                graph.operations[created.name] = created

            graph.insert_op_before(A=created, B=op, input_idx=var_index)
            new_var = created.outputs[0]
            new_var.shape = shape
            new_var.is_parameter = var.is_parameter
            new_var.dtype = var.dtype

            shape_param = graph.create_variable(value=torch.Tensor(shape).to(torch.int64), is_parameter=True, dest_ops=[created])
            shape_param.dtype = DataType.INT64

        else:
            raise ValueError(f"Unexpected Error in insert reshape node, var: {var.name}, Op: {op.name}.")

        return created


    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type != "MatMul" or not isinstance(op, QuantableOperation):
            return op

        input0 = op.inputs[0]
        input1 = op.inputs[1]
        input1_config = op.input_quant_config[1]
        input1_num_of_bits = input1_config.num_of_bits
        input1_n_size = input1.shape[-1]
        if input0.shape is None or input1.shape is None:
            logger.error("input shape is None")
            return op

        if input1_num_of_bits != 8 and input1_num_of_bits != 16:
            logger.warning(f"The num_of_bits of input1 {input1_num_of_bits} is not supported.")
            return op

        input0_dims = len(input0.shape)
        input1_dims = len(input1.shape)
        input1_orig_shape = input1.shape
        align = 16 if input1_num_of_bits == 8 else 8

        if input0_dims == 1 and input1_dims == 1:
            # Don't need to insert anything.
            return op

        if input1_n_size % align == 0:
            if input1_dims == 2:
                # CN -> (N/align)C(align) = (N/align)HWC(align)
                c, n = input1.shape
                InsertPreNodeOfMatMulPattern.insert_reshape_node(graph = graph, 
                                                                var = op.inputs[1],
                                                                op = op, 
                                                                shape = [c, n // align, align])
                InsertPreNodeOfMatMulPattern.insert_transpose_node(graph = graph, 
                                                                var = op.inputs[1],
                                                                op = op, 
                                                                perm = [1, 0, 2])
                InsertPreNodeOfMatMulPattern.insert_reshape_node(graph = graph, 
                                                                var = op.inputs[1],
                                                                op = op, 
                                                                shape = input1_orig_shape)
            elif input1_dims > 2:
                c, n = input1.shape[-2:]
                InsertPreNodeOfMatMulPattern.insert_reshape_node(graph = graph, 
                                                                var = op.inputs[1],
                                                                op = op, 
                                                                shape = [*input1.shape[: -2], c, n // align, align])
                InsertPreNodeOfMatMulPattern.insert_transpose_node(graph = graph, 
                                                                var = op.inputs[1],
                                                                op = op, 
                                                                perm = list(range(len(input1.shape) - 2)) + [-2, -3, -1])
                InsertPreNodeOfMatMulPattern.insert_reshape_node(graph = graph, 
                                                                var = op.inputs[1],
                                                                op = op, 
                                                                shape = input1_orig_shape)

        return op


class FuseReluLikePattern(OperationExporter):
    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if not isinstance(op, QuantableOperation):
            return op

        # The FUSE_OP_PATTERNS may remove some ops.
        if op.name not in graph.operations:
            return op

        if op.type in ["Conv", "Gemm"]:
            op.attributes["activation"] = "Linear"
            downstream_op = graph.get_downstream_operations(op)
            if (
                len(downstream_op) == 1
            ):  # the downstream op have only one op and this op is relu
                # if downstream_op[0].type in ["Relu", "Clip"]:
                if downstream_op[0].type in ["Relu"]:
                    logger.debug(
                        f"fuse {op.type}:{op.name} and {downstream_op[0].type}:{downstream_op[0].name}"
                    )
                    conv_quant_config = op.config
                    relu_quant_config = downstream_op[0].config
                    new_config = OperationQuantizationConfig(
                        conv_quant_config.input_quantization_config,
                        relu_quant_config.output_quantization_config,
                    )

                    # graph.remove_operation(downstream_op[0], keep_coherence=True)
                    graph = fuse_downstream_operation(
                        graph, downstream_op[0], keep_coherence=True
                    )
                    op.config = new_config
                    op.attributes["activation"] = downstream_op[0].type

        return op


class QuantVariableToIntPattern(OperationExporter):
    @staticmethod
    def calculate_exponent(config: TensorQuantizationConfig):
        if not config.policy.has_property(QuantizationProperty.LINEAR):
            raise ValueError("Critical Quantization Error! Non-linear config detected.")
        if config.policy.has_property(QuantizationProperty.ASYMMETRICAL):
            raise ValueError(
                "Critical Quantization Error! Asymmetrical config detected."
            )

        if not config.scale:
            return None

        exponent = None
        if config.policy.has_property(
            QuantizationProperty.PER_TENSOR
        ) and config.policy.has_property(QuantizationProperty.POWER_OF_2):
            scale = convert_any_to_numpy(config.scale)
            exponent = [int(np.log2(scale))]
        elif config.policy.has_property(
            QuantizationProperty.PER_CHANNEL
        ) and config.policy.has_property(QuantizationProperty.POWER_OF_2):
            scale = convert_any_to_numpy(config.scale)
            exponent = np.log2(scale).astype(int)
        return exponent

    def export(self, op: QuantableOperation, graph: BaseGraph, **kwargs) -> Operation:
        if not isinstance(op, QuantableOperation):
            logger.info("skip not QuantableOperation")
            return op

        # collect quantable vars, where we need to quantize parameters
        info = ExporterPatternInfo()

        for config, var in [_ for _ in op.config_with_variable]:
            if not var or not config:
                logger.info("skip not config or var")
                continue

            if not EspdlQuantHelper.TQC_Exportable_Check(TQC=config, bounded_var=var):
                continue

            if not info.get_var_config(var.name):
                info.add_var_config(var.name, config)

            if not info.get_var_exponents(var.name):
                exponent = QuantVariableToIntPattern.calculate_exponent(config)
                if exponent:
                    info.add_var_exponents(var.name, exponent)
                    info.add_var_config(var.name, config)
                    logger.debug(f"{var.name} exponent: {exponent}")
                else:
                    logger.info(
                        "Skip %s from (op name:%s, type:%s) because it's not quantized"
                        % (var.name, op.name, op.type)
                    )
            else:
                continue

            if var.is_parameter:
                assert len(var.dest_ops) == 1, (
                    f"Can not export variable {var.name}, cause it has more than 1 destination operations. "
                    "PPQ require all parameters to have only 1 destination operation."
                )

                # override quantization state, so that we can export parameter correctly.
                if config.state == QuantizationStates.BAKED:
                    config.state = QuantizationStates.ACTIVATED
                if config.state == QuantizationStates.PASSIVE_BAKED:
                    config.state = QuantizationStates.PASSIVE

                if config.policy.has_property(QuantizationProperty.LINEAR):
                    var.value = PPQLinearQuant_toInt(tensor=var.value, config=config)
            elif not var.is_parameter:
                if config.policy.has_property(QuantizationProperty.LINEAR):
                    quant_type = op.attributes.get("quant_type", None)
                    if quant_type == EspQuantType.S8:
                        var.dtype = DataType.INT8
                    elif quant_type == EspQuantType.S16:
                        var.dtype = DataType.INT16
                    else:
                        var.dtype = DataType.FP32

        return op


class ResetParamLayoutPattern(OperationExporter):
    def reset_conv_filter_layout(self, tensor, quant_type, group=None):
        if len(tensor.shape) != 4:
            logger.error(
                f"Conv filter should be 4D tensor, but got {len(tensor.shape)}D tensor."
            )
            return tensor, LayoutAnnotation.NCHW

        n, c, h, w = (
            tensor.shape
        )  # n denotes output channels, c denotes input channels,
        tensor = tensor.permute(0, 2, 3, 1)  # NCHW -> NHWC

        align = 16 if quant_type == EspQuantType.S8 else 8
        aligned_len = n // align * align
        aligned_tensor = tensor[0:aligned_len, ...]
        aligned_tensor = aligned_tensor.reshape(
            n // align, align, h, w, c
        )  # NHWC -> (N/align,align)HWC
        # (N/align,align)HWC -> (N/align)HWC(align)
        aligned_tensor = aligned_tensor.permute(0, 2, 3, 4, 1)
        # (N/align)HWC(align) -> (aligned_len)HWC
        aligned_tensor = aligned_tensor.reshape(aligned_len, h, w, c)

        if n % align != 0:
            unaligned_tensor = tensor[aligned_len:n, ...]  # NHWC
            if group == 1 or group == None:
                aligned_tensor = torch.cat((aligned_tensor, unaligned_tensor), 0)
            else:
                n_remain = n - aligned_len
                unaligned_tensor = unaligned_tensor.permute(
                    3, 1, 2, 0
                )  # depthwise unaligned: NHWC -> CHWN
                unaligned_tensor = unaligned_tensor.reshape(n_remain, h, w, c)
                aligned_tensor = torch.cat((aligned_tensor, unaligned_tensor), 0)

            if align == 16:
                layout = LayoutAnnotation.N16HWC16_UNALIGNED
            else:
                layout = LayoutAnnotation.N8HWC8_UNALIGNED
        else:
            if align == 16:
                layout = LayoutAnnotation.N16HWC16
            else:
                layout = LayoutAnnotation.N8HWC8

        # TODO:: modify the layout of depthwise conv in ESP-DL, keep it same with conv
        if group == 1 or group == None:
            aligned_tensor = aligned_tensor.reshape(h, w, c, n)  # reshape to HWCN
        else:
            aligned_tensor = aligned_tensor.reshape(h, w, n, c)  # reshape to HWNC
        return aligned_tensor, layout

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        quant_type = op.attributes.get("quant_type", None)
        if (
            quant_type == None
            or quant_type == EspQuantType.F32
            or not isinstance(op, QuantableOperation)
        ):
            return op

        info = ExporterPatternInfo()

        if op.type == "Conv":
            for var in op.inputs:
                if not var.is_parameter:
                    continue

                tensor = var.value
                if len(tensor.shape) == 4:  # Conv2d Filter
                    group = op.attributes.get("group", None)
                    aligned_tensor, layout = self.reset_conv_filter_layout(
                        tensor, quant_type, group
                    )
                    info.add_var_layout(var.name, layout)
                    var.value = aligned_tensor
                    logger.debug(
                        f"reset {op.type}:{op.name}, shape:{tensor.shape}, layout to {layout}"
                    )

        elif op.type == "Gemm":
            for var in op.inputs:
                if not var.is_parameter:
                    continue

                # fix the transB attribute is 0
                tensor = var.value
                alpha = op.attributes.get("alpha", 1.0)
                beta = op.attributes.get("beta", 1.0)
                assert (
                    alpha == 1.0 and beta == 1.0
                ), "alpha and beta must be 1.0 and 0.0"

                if len(tensor.shape) == 2:  # Gemm Filter
                    trans_filter = op.attributes.get("transB", 0)
                    if trans_filter != 0:
                        logger.debug(
                            "transB is not 0, transpose the filter and reset transB"
                        )
                        op.attributes["transB"] = 0  # update 'transB'
                        tensor = tensor.transpose(1, 0)  # [N, C] -> [C, N]
                    tensor = tensor.unsqueeze(-1).unsqueeze(-1)  # CN -> CNHW
                    # CNHW -> NCHW, same with conv2d filter
                    tensor = tensor.permute(1, 0, 2, 3)

                    aligned_tensor, layout = self.reset_conv_filter_layout(
                        tensor, quant_type, None
                    )
                    info.add_var_layout(var.name, layout)
                    var.value = aligned_tensor
                    logger.debug(
                        f"reset {op.type}:{op.name}, shape:{var.value.shape}, layout to {layout}"
                    )

        return op


class AddLUTPattern(OperationExporter):
    def __init__(self, int16_step=1) -> None:
        super().__init__()
        self.int16_step = int(int16_step)  # the step of int16 LUT

    def get_scale(self, var: Variable, info: ExporterPatternInfo) -> torch.Tensor:
        exponent = info.get_var_exponents(var.name)
        scale = 1.0
        if exponent:
            if isinstance(exponent, list):
                scale = 2 ** exponent[0]
            else:
                scale = 2**exponent

        return scale

    def calculate_lut(
        self,
        op: QuantableOperation,
        info: ExporterPatternInfo,
        max: int,
        min: int,
        step: int = 1,
    ) -> torch.Tensor:
        # Get forward function
        platform_dispatching_table = OPERATION_FORWARD_TABLE[op.platform]
        if op.type not in platform_dispatching_table:
            raise NotImplementedError(
                f"Graph op: {op.name}({op.type}) "
                f"has no backend implementation on target platform {op.platform}. "
                "Register this op to ppq.executor.base.py and ppq.executor.op first"
            )
        operation_forward_func = platform_dispatching_table[op.type]

        # Calculate output and lut
        input = torch.arange(min, max + 1, step=step, dtype=torch.float)
        input = input * self.get_scale(op.inputs[0], info)
        inputs = [input]

        if len(op.inputs) > 1:
            for op_input in op.inputs[1:]:
                inputs.append(op_input.value * self.get_scale(op_input, info))
        output = operation_forward_func(op, inputs)
        lut = PPQLinearQuant_toInt(output, op.output_quant_config[0])

        return lut

    def get_lut_name(self, op: Operation, info: ExporterPatternInfo):
        index = len(info.luts)
        name = f"{op.type}_lut_{index}"
        return name

    def check_op(self, op: Operation):
        """
        True if this op can be implemented by LUT, otherwise False
        """

        if op.type == "PRelu":
            if op.inputs[1].value.numel() == 1:
                return True
            else:
                return False
        elif len(op.outputs) > 1 or len(op.inputs) > 1:
            return False
        elif op.type in ACTIVATION_OP_SET or op.type in MATH_OP_SET:
            return True

        return False

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        quant_type = op.attributes.get("quant_type", None)
        if (
            quant_type == None
            or quant_type == EspQuantType.F32
            or not isinstance(op, QuantableOperation)
        ):
            return op

        info = ExporterPatternInfo()

        if self.check_op(op):
            lut = None
            if quant_type == EspQuantType.S8:
                lut = self.calculate_lut(op, info, 127, -128, 1)
            elif quant_type == EspQuantType.S16 and self.int16_step > 0:
                lut = self.calculate_lut(op, info, 2**15 - 1, - 2**15, self.int16_step)

            if lut != None:
                lut_name = self.get_lut_name(op, info)
                op.attributes["lut"] = lut_name
                info.add_lut(lut_name, lut, info.get_var_exponents(op.outputs[0].name))

        return op



from collections import defaultdict

import numpy as np
from ppq.log import NaiveLogger

from . import ppl_caffe_pb2

logger = NaiveLogger.get_logger('PPQ')

def de_inplace(net_def: ppl_caffe_pb2.NetParameter) -> ppl_caffe_pb2.NetParameter:
    """Remove inplace layer in netdef If the names of bottom and top are same,
    it means the computation of this layer is in place."""

    def new_name(_name):
        if current_write_times[_name] == total_write_times[_name]:
            return _name
        else:
            return f'{_name}_ver{current_write_times[_name]}'

    total_write_times = {}
    for layer in net_def.layer:
        for top in layer.top:
            total_write_times.setdefault(top, 0)
            total_write_times[top] += 1

    current_write_times = {}
    for name in net_def.input:
        # Init input var, only read in the beginning
        total_write_times[name] = 0
        current_write_times[name] = 0

    for layer in net_def.layer:
        for i, name in enumerate(layer.bottom):
            layer.bottom[i] = new_name(name)
        for i, name in enumerate(layer.top):
            current_write_times.setdefault(name, 0)
            current_write_times[name] += 1
            layer.top[i] = new_name(name)

    return net_def

def merge_batchnorm_scale(caffe_net: ppl_caffe_pb2.NetParameter) -> ppl_caffe_pb2.NetParameter:
    new_net = ppl_caffe_pb2.NetParameter()
    new_net.CopyFrom(caffe_net)
    new_net.ClearField('layer')

    idx = 0
    while idx < len(caffe_net.layer):
        if idx != len(caffe_net.layer) - 1 and caffe_net.layer[idx].type == 'BatchNorm' and caffe_net.layer[idx + 1].type == 'Scale':
            batchnorm_layer = caffe_net.layer[idx]
            scale_layer = caffe_net.layer[idx + 1]
            if not batchnorm_layer.batch_norm_param.use_global_stats:
                logger.warning('Now cannot convert BatchNorm layer with use_global_stats==False. Change it to True')
            # Generate BN layer from batchnorm + scale
            layer = ppl_caffe_pb2.LayerParameter(type='BN', name=batchnorm_layer.name)
            layer.bottom[:] = batchnorm_layer.bottom[:]
            layer.top[:] = scale_layer.top[:]
            # Set attribute
            layer.bn_param.var_eps = batchnorm_layer.batch_norm_param.eps
            # assign blobs in order: weight, bias, mean, var
            weight_blob = ppl_caffe_pb2.BlobProto()
            weight_blob.CopyFrom(scale_layer.blobs[0])
            layer.blobs.extend([weight_blob])

            bias_blob = ppl_caffe_pb2.BlobProto()
            if scale_layer.scale_param.bias_term:
                bias_blob.CopyFrom(scale_layer.blobs[1])
                layer.blobs.extend([bias_blob])
            else:
                # bias = 0
                bias_blob.data.extend(np.zeros(len(weight_blob.data), dtype='f'))
                layer.blobs.extend([bias_blob])

            mean_blob = ppl_caffe_pb2.BlobProto()
            mean_blob.CopyFrom(batchnorm_layer.blobs[0])
            layer.blobs.extend([mean_blob])
            var_blob = ppl_caffe_pb2.BlobProto()
            var_blob.CopyFrom(batchnorm_layer.blobs[1])
            layer.blobs.extend([var_blob])
            # Update running_mean and running_var according to scale_factor
            if len(batchnorm_layer.blobs) == 3:
                factor = np.array(batchnorm_layer.blobs[2].data, dtype='f')
                layer.blobs[2].ClearField('data')
                running_mean = np.array(batchnorm_layer.blobs[0].data, dtype='f') / factor
                layer.blobs[2].data.extend(running_mean)
                layer.blobs[3].ClearField('data')
                running_var = np.array(batchnorm_layer.blobs[1].data, dtype='f') / factor
                layer.blobs[3].data.extend(running_var)

            new_net.layer.extend([layer])
            idx += 2
        else:
            layer = ppl_caffe_pb2.LayerParameter()
            layer.CopyFrom(caffe_net.layer[idx])
            new_net.layer.extend([layer])
            idx += 1
    return new_net

def optimize_for_export(caffe_net: ppl_caffe_pb2.NetParameter) -> ppl_caffe_pb2.NetParameter:
    """Simplify some caffe ops Pattern 1: combine multi-slices to one Pattern
    2: combine multi-eltwise back to one."""
    slice_combine(caffe_net)
    eltwise_combine(caffe_net)
    return caffe_net


def slice_combine(caffe_net: ppl_caffe_pb2.NetParameter):
    slice_layer = [layer for layer in caffe_net.layer if layer.type == 'Slice']
    slice_opt_set = []
    matched = []
    # Get all slice_set which can be merged together
    for i in range(len(slice_layer)):
        if slice_layer[i] in matched:
            continue
        else:
            slice_set = [slice_layer[i]]
        for j in range(i + 1, len(slice_layer)):
            if slice_layer[j] not in matched:
                same_flag = (slice_layer[i].bottom == slice_layer[j].bottom) and \
                            (slice_layer[i].slice_param.axis == slice_layer[j].slice_param.axis)
                if same_flag:
                    slice_set.append(slice_layer[j])
        if len(slice_set) > 1:
            slice_opt_set.append(slice_set)
            matched.extend(slice_set)
    # Merge Slice layer
    for layer_set in slice_opt_set:
        remained_layer = layer_set[0]
        slice_point = [layer.slice_param.slice_point for layer in layer_set] + [[0]]
        slice_point = sorted(set([item for sublist in slice_point for item in sublist]))

        name_dict = defaultdict(list)
        # An example: name_dict = {slice_point_0: ['337', '338_front'], slice_point_1: ['338', '337_behind']}
        for layer in layer_set:
            points = layer.slice_param.slice_point
            cuttings = [slice_point[slice_point.index(points[0]) - 1]] + list(points)
            for name, cut_point in zip(layer.top, cuttings):
                name_dict[cut_point].append(name)

        changed_name_dict = {}
        # An example: changed_name_dict = {'338_front': '337', '337_behind': '338'}
        new_top = []
        for names in name_dict.values():
            key = [name for name in names if ('_front' not in name) and ('_behind' not in name)]
            if not key:
                key = names[0]
                logger.warning(f'Cannot find origin name in slice combination. Use {key}')
            else:
                assert len(key) == 1, 'Find multiple origin names in slice combination. Only support one.'
                key = key[0]
            new_top.append(key)
            names.remove(key)
            changed_name_dict.update([(i, key) for i in names])

        remained_layer.top[:] = new_top
        remained_layer.slice_param.slice_point[:] = slice_point[1:]  # remove 0 in slice_point

        # update changed name
        for layer in caffe_net.layer:
            for i, item in enumerate(layer.bottom):
                if item in changed_name_dict:
                    layer.bottom[i] = changed_name_dict[item]
        # delete other slice layer
        for i in range(1, len(layer_set)):
            caffe_net.layer.remove(layer_set[i])


def eltwise_combine(caffe_net: ppl_caffe_pb2.NetParameter):
    eltwise_layer = [layer for layer in caffe_net.layer if layer.type == 'Eltwise']
    # Step 1: Combine mul op generated from coeff != 1.0 in origin caffe eltwise op
    eltwise_opt_set = []
    matched = []
    # Get all eltwise_set which can be merged together
    for i in range(len(eltwise_layer)):
        if '_mul' not in eltwise_layer[i].name:
            eltwise_set = [eltwise_layer[i]]
            find_out = eltwise_layer[i].bottom
        else:
            continue

        for layer in eltwise_layer:
            if layer not in matched:
                # Eltwise only has 1 output
                same_flag = ('_mul' in layer.name) and (layer.top[0] in find_out)
                if same_flag:
                    eltwise_set.append(layer)

        if len(eltwise_set) > 1:
            eltwise_opt_set.append(eltwise_set)
            matched.extend(eltwise_set)
    # Merge mul op
    for eltwise_set in eltwise_opt_set:
        remained_layer = [layer for layer in eltwise_set if '_mul' not in layer.name]
        assert len(remained_layer) == 1, f'Cannot find root layer for {eltwise_set[0]}'
        remained_layer = remained_layer[0]
        eltwise_set.remove(remained_layer)

        new_bottom = {}
        for layer in eltwise_set:
            assert len(layer.bottom) == len(layer.top) == 1
            new_bottom[layer.top[0]] = (layer.bottom[0], layer.eltwise_param.coeff[0])
            # remove layer
            eltwise_layer.remove(layer)
            caffe_net.layer.remove(layer)

        if len(remained_layer.eltwise_param.coeff) == 0:
            remained_layer.eltwise_param.coeff[:] = [1.0] * len(remained_layer.bottom)
        for i, item in enumerate(remained_layer.bottom):
            if item in new_bottom:
                remained_layer.bottom[i] = new_bottom[item][0]
                remained_layer.eltwise_param.coeff[i] = new_bottom[item][1]

    # Step 2:Combine binary op to multi-input eltwise
    eltwise_opt_set = []
    matched = []
    # Get all eltwise_set which can be merged together
    for layer_target in eltwise_layer:
        eltwise_set = [layer_target]
        for layer_find in eltwise_layer:
            # the layer_target must only have one successor
            same_flag = (layer_target.eltwise_param.operation == layer_find.eltwise_param.operation) and \
                (layer_target.top[0] in layer_find.bottom) and \
                len([layer for layer in caffe_net.layer if layer_target.top[0] in layer.bottom]) == 1
            if same_flag:
                eltwise_set.append(layer_find)

        if len(eltwise_set) == 2:
            eltwise_opt_set.append(eltwise_set)
            matched.extend(eltwise_set)
    # Merge binary op
    for eltwise_set in reversed(eltwise_opt_set):
        remained_layer, combine_layer = eltwise_set
        if len(remained_layer.eltwise_param.coeff) == 0:
            remained_layer.eltwise_param.coeff[:] = [1.0] * len(remained_layer.bottom)
        if len(combine_layer.eltwise_param.coeff) == 0:
            combine_layer.eltwise_param.coeff[:] = [1.0] * len(remained_layer.bottom)

        idx = list(combine_layer.bottom).index(remained_layer.top[0])
        if combine_layer.eltwise_param.coeff[idx] != 1.0:
            continue
        else:
            combine_layer.bottom.pop(idx)
            combine_layer.eltwise_param.coeff.pop(idx)
            remained_layer.bottom.extend(combine_layer.bottom)
            remained_layer.eltwise_param.coeff.extend(combine_layer.eltwise_param.coeff)
            remained_layer.top[:] = combine_layer.top[:]
            caffe_net.layer.remove(combine_layer)



from typing import Iterable

import torch
import torchvision
from torch.utils.data import DataLoader

from ppq import BaseGraph, QuantizationSettingFactory, TargetPlatform
from ppq.api import export_ppq_graph, quantize_torch_model

BATCHSIZE = 32
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cuda' # only cuda is fully tested :(  For other executing device there might be bugs.
PLATFORM = TargetPlatform.PPL_CUDA_INT8  # identify a target platform for your network.

def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]

def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)

# Load a pretrained mobilenet v2 model
model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# create a setting for quantizing your network with PPL CUDA.
quant_setting = QuantizationSettingFactory.pplcuda_setting()
quant_setting.equalization = True # use layerwise equalization algorithm.
quant_setting.dispatcher   = 'conservative' # dispatch this network in conservertive way.

# Load training data for creating a calibration dataloader.
calibration_dataset = load_calibration_dataset()
calibration_dataloader = DataLoader(
    dataset=calibration_dataset,
    batch_size=BATCHSIZE, shuffle=True)

# quantize your model.
quantized = quantize_torch_model(
    model=model, calib_dataloader=calibration_dataloader,
    calib_steps=32, input_shape=[BATCHSIZE] + INPUT_SHAPE,
    setting=quant_setting, collate_fn=collate_fn, platform=PLATFORM,
    onnx_export_file='Output/onnx.model', device=DEVICE, verbose=0)

# Quantization Result is a PPQ BaseGraph instance.
assert isinstance(quantized, BaseGraph)

# export quantized graph.
export_ppq_graph(graph=quantized, platform=PLATFORM,
                 graph_save_to='Output/quantized(onnx).onnx',
                 config_save_to='Output/quantized(onnx).json')



from .AscendQuantizer import AscendQuantizer
from .base import BaseQuantizer
from .DSPQuantizer import PPL_DSP_Quantizer, PPL_DSP_TI_Quantizer
from .EspdlQuantizer import EspdlInt16Quantizer, EspdlQuantizer, EspdlS3Quantizer, EspdlS3Int16Quantizer
from .FP8Quantizer import GraphCoreQuantizer, TensorRTQuantizer_FP8

# from .TRTQuantizer import TensorRTQuantizer
from .FPGAQuantizer import FPGAQuantizer
from .MetaxQuantizer import MetaxChannelwiseQuantizer, MetaxTensorwiseQuantizer
from .MNNQuantizer import MNNQuantizer
from .MyQuantizer import ExtQuantizer
from .NCNNQuantizer import NCNNQuantizer
from .NXPQuantizer import NXP_Quantizer
from .OpenvinoQuantizer import OpenvinoQuantizer
from .ORTQuantizer import OnnxruntimeQuantizer
from .PPLQuantizer import PPLCUDAQuantizer
from .RKNNQuantizer import RKNN_PerChannelQuantizer, RKNN_PerTensorQuantizer
from .TengineQuantizer import TengineQuantizer
from .TensorRTQuantizer import TensorRTQuantizer, TensorRTQuantizer_InputOnly



import json

from ppq.core import (DataType, QuantizationProperty, QuantizationStates,
                      TargetPlatform, TensorQuantizationConfig)
from ppq.IR import BaseGraph
from ppq.IR.quantize import QuantableOperation

from .onnx_exporter import OnnxExporter
from .util import convert_value


def convert_type(platform: TargetPlatform) -> str:
    if platform == TargetPlatform.PPL_CUDA_INT8: return 'INT8'
    if platform == TargetPlatform.SOI: return None
    if platform == TargetPlatform.FP32: return None
    raise TypeError(f'Unsupported platform type. ({str(platform)})')


class PPLBackendExporter(OnnxExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        var_quant_info_recorder, op_platform_recorder = {}, {}
        for operation in graph.operations.values():
            if not isinstance(operation, QuantableOperation): continue
            for config, var in operation.config_with_variable:
                if not config.can_export(): continue

                # PATCH 2021.11.25
                # REMOVE BIAS FROM CONFIGURATION
                if config.num_of_bits > 8: continue

                if config.state in {
                    QuantizationStates.FP32,
                }: continue
                # Simply override recorder is acceptable here,
                # we do not support mix precision quantization for CUDA backend now.
                # All configurations for this variable should keep identical towards each other.

                if config.state == QuantizationStates.PASSIVE and var.name in var_quant_info_recorder: continue
                var_quant_info_recorder[var.name] = config

        # ready to render config to json.
        for var in var_quant_info_recorder:
            config = var_quant_info_recorder[var]
            assert isinstance(config, TensorQuantizationConfig)
            tensorwise = config.policy.has_property(QuantizationProperty.PER_TENSOR)
            var_quant_info_recorder[var] = {
                'bit_width'  : config.num_of_bits,
                'per_channel': config.policy.has_property(QuantizationProperty.PER_CHANNEL),
                'quant_flag' : True,
                'sym'        : config.policy.has_property(QuantizationProperty.SYMMETRICAL),
                'scale'      : convert_value(config.scale, tensorwise, DataType.FP32),
                'zero_point' : convert_value(config.offset, tensorwise, DataType.INT32),
                'tensor_min' : convert_value(config.scale * (config.quant_min - config.offset), tensorwise, DataType.FP32),
                'tensor_max' : convert_value(config.scale * (config.quant_max - config.offset), tensorwise, DataType.FP32),
                'q_min'      : config.quant_min,
                'q_max'      : config.quant_max,
                'hash'       : config.__hash__(),
                'dominator'  : config.dominated_by.__hash__()
            }

        for op in graph.operations.values():
            if convert_type(op.platform) is not None:
                op_platform_recorder[op.name] = {
                    'data_type': convert_type(op.platform)
                }

        exports = {
            'quant_info': var_quant_info_recorder,
            'op_info': op_platform_recorder}

        with open(file=config_path, mode='w') as file:
            json.dump(exports, file, indent=4)



from .onnxruntime_exporter import ONNXRUNTIMExporter
import torch
from ppq.core import (QuantizationProperty,
                      TensorQuantizationConfig, 
                      convert_any_to_torch_tensor)
from ppq.IR import (BaseGraph, Operation,
                    Variable)
from ppq.utils.round import ppq_tensor_round


class OpenvinoExporter(ONNXRUNTIMExporter):
    """Openvino 喜欢所有 QuantizeLinear 上面都带 axis 属性，所以这里需要单独处理一下"""
    
    def insert_quantize_node(
        self, graph: BaseGraph, 
        var: Variable, config: TensorQuantizationConfig, 
        op: Operation) -> Operation:
        """
        Insert a Quantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of Quantize Node: QuantizeLinear and QuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> INT
            offset_dtype, value_type = self.infer_qtype(config)
            scale  = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type='QuantizeLinear', attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else: created.attributes['axis'] = 0
            
            if var in op.inputs:  graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs: graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else: raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].dtype = value_type
            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created
        
        elif config.policy.has_property(QuantizationProperty.FLOATING):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> FP8
            scale  = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_torch_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type='QuantizeFloating', 
                attributes={
                    'min': config.quant_min, 
                    'max': config.quant_max, 
                    'exponent': config.exponent_bits, 
                    'mantissa': config.mantissa_bits})

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else: created.attributes['axis'] = None
            
            if var in op.inputs:  graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs: graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else: raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created
        
        else:
            raise TypeError(
                f'PPQ Can not export quantization information with variable {var.name}, '
                'Unexpected Quantization property.')

    def insert_dequantize_node(
        self, graph: BaseGraph, 
        var: Variable, config: TensorQuantizationConfig, 
        op: Operation) -> Operation:
        """
        Insert a DeQuantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of DeQuantize Node: DeQuantizeLinear and DeQuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            offset_dtype, value_type = self.infer_qtype(config)
            scale  = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)
            
            created = graph.create_operation(op_type='DequantizeLinear', attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else: created.attributes['axis'] = 0
            
            if var in op.inputs:  graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs: graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else: raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')
            
            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.inputs[0].dtype = value_type
            created.inputs[0].shape = var.shape
            created.outputs[0].shape = var.shape
            created.outputs[0].dtype = torch.float32

            return created

        elif config.policy.has_property(QuantizationProperty.FLOATING):
            scale  = convert_any_to_torch_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_torch_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type='DequantizeFloating', 
                attributes={
                    'min': config.quant_min, 
                    'max': config.quant_max, 
                    'exponent': config.exponent_bits, 
                    'mantissa': config.mantissa_bits})

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes['axis'] = config.channel_axis
            else: created.attributes['axis'] = None
            
            if var in op.inputs:  graph.insert_op_before(A=created, B=op, input_idx=op.inputs.index(var))
            elif var in op.outputs: graph.insert_op_after(A=created, B=op, output_idx=op.outputs.index(var))
            else: raise ValueError(f'Unexpected Error in Exporting Op {op.name}({op.type}).')

            graph.create_variable(name=None, value=scale, is_parameter=True, dest_ops=[created])
            graph.create_variable(name=None, value=offset, is_parameter=True, dest_ops=[created])

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape

            return created

        else:

            raise TypeError(
                f'PPQ Can not export quantization information with variable {var.name}, '
                'Unexpected Quantization property.')



import json
import os
from copy import deepcopy
from typing import List

import torch
from ppq.core import (PPQ_CONFIG, DataType, QuantizationProperty,
                      QuantizationStates, TargetPlatform,
                      TensorQuantizationConfig, convert_any_to_torch_tensor)
from ppq.executor.torch import TorchExecutor
from ppq.IR import BaseGraph, GraphExporter
from ppq.IR.quantize import QuantableOperation, QuantableVariable
from ppq.log import NaiveLogger

from .caffe import ppl_caffe_pb2
from .caffe.caffe_export_utils import CaffeOpExporter, caffe_export_map
from .caffe.caffe_graph_optim import optimize_for_export
from .util import convert_value

logger = NaiveLogger.get_logger('PPQ')

def convert_type(platform: TargetPlatform) -> str:
    if platform == TargetPlatform.PPL_CUDA_INT8: return 'INT8'
    if platform == TargetPlatform.PPL_DSP_INT8: return 'INT8'
    if platform == TargetPlatform.SOI: return None
    if platform == TargetPlatform.FP32: return None
    raise TypeError(f'Unsupported platform type. ({str(platform)})')

OPEARTION_PLATFORM_FORMAT = \
"""
{name}:(
    platform:\t{platform},
)
"""

class CaffeExporter(GraphExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        var_quant_info_recorder, op_platform_recorder = {}, {}
        for operation in graph.operations.values():
            if not isinstance(operation, QuantableOperation): continue
            for config, var in operation.config_with_variable:
                if not config.can_export(): continue

                # PATCH 2021.11.25
                # REMOVE BIAS FROM CONFIGURATION
                if config.num_of_bits > 8:
                        continue

                if config.state in {
                    QuantizationStates.FP32,
                }: continue
                # Simply override recorder is acceptable here,
                # we do not support mix precision quantization for CUDA backend now.
                # All configurations for this variable should keep identical towards each other.

                if config.state == QuantizationStates.PASSIVE and var.name in var_quant_info_recorder: continue
                var_quant_info_recorder[var.name] = config

        # ready to render config to json.
        for var in var_quant_info_recorder:
            config = var_quant_info_recorder[var]
            assert isinstance(config, TensorQuantizationConfig)
            tensorwise = config.policy.has_property(QuantizationProperty.PER_TENSOR)
            var_quant_info_recorder[var] = {
                'bit_width'  : config.num_of_bits,
                'per_channel': config.policy.has_property(QuantizationProperty.PER_CHANNEL),
                'quant_flag' : True,
                'sym'        : config.policy.has_property(QuantizationProperty.SYMMETRICAL),
                'scale'      : convert_value(config.scale, tensorwise, DataType.FP32),
                'zero_point' : convert_value(config.offset, tensorwise, DataType.INT32),
                'tensor_min' : convert_value(config.scale * (config.quant_min - config.offset), tensorwise, DataType.FP32),
                'tensor_max' : convert_value(config.scale * (config.quant_max - config.offset), tensorwise, DataType.FP32),
                'q_min'      : config.quant_min,
                'q_max'      : config.quant_max,
                'hash'       : config.__hash__(),
                'dominator'  : config.dominated_by.__hash__()
            }

        for op in graph.operations.values():
            if convert_type(op.platform) is not None:
                op_platform_recorder[op.name] = {
                    'data_type': convert_type(op.platform)
                }

        exports = {
            'quant_info': var_quant_info_recorder,
            'op_info': op_platform_recorder}

        with open(file=config_path, mode='w') as file:
            json.dump(exports, file, indent=4)

    def prepare_model(self, graph: BaseGraph, input_shapes: List[List[int]]):
        # trace model for exporting.
        device = 'cuda' if torch.cuda.is_available() else 'cpu'
        if all([var.value is not None for var in graph.inputs.values()]):
            inputs = {var.name: convert_any_to_torch_tensor(var.value).to(device) for var in graph.inputs.values()}
        elif all([var.shape is not None for var in graph.inputs.values()]):
            inputs = {var.name: torch.randn(*var.shape, dtype=DataType.to_torch(var.dtype), device=device)\
                for var in graph.inputs.values()}
        else:
            assert len([input_shapes]) == len(graph.inputs), (
                'must provide equal number of input shapes for caffe export without quantization')
            # assume all fp32 type, because that's the usual case
            inputs = [torch.randn(*shape, device=device) for shape in input_shapes]
        tracer = TorchExecutor(graph, device=device)
        tracer.tracing_operation_meta(inputs, list(graph.outputs.keys()))

        # build caffe protobuf
        caffe_model = ppl_caffe_pb2.NetParameter()
        caffe_model.name = graph._name if graph._name else 'PPQ Exported Caffe Model'

        # add caffe input info
        for name, var in graph.inputs.items():
            caffe_model.input.append(name)
            input_shape = ppl_caffe_pb2.BlobShape()
            var.shape[0] = 1
            input_shape.dim.extend(var.shape)
            caffe_model.input_shape.extend([input_shape])

        # export op
        for op in graph.topological_sort():
            if op.type not in caffe_export_map:
                raise NotImplementedError(
                    f'{op.type} converted to Caffe OP is not supported in PPQ export parser yet')

            caffe_op = caffe_export_map[op.type](op)
            assert isinstance(caffe_op, CaffeOpExporter)

            layer = caffe_op.parse()
            if not isinstance(layer, (list, tuple)):
                layer = [layer]
            caffe_model.layer.extend(layer)

        caffe_model = optimize_for_export(caffe_model)
        caffe_proto = deepcopy(caffe_model)
        for layer in caffe_proto.layer: del layer.blobs[:]

        return caffe_model, caffe_proto

    def dump_to_file(self, caffe_model: ppl_caffe_pb2.NetParameter,
                     caffe_proto: ppl_caffe_pb2.NetParameter, file_path: str):

        # it's stupid but for keeping api the same
        # 就是说 caffemodel 总是要保存两个东西出来，但是我们的 api 只留了一个地方
        # 所以我们就只能这样了
        prefix, _       = os.path.splitext(file_path)
        prototxt_path   = prefix + '.prototxt'
        caffemodel_path = prefix + '.caffemodel'

        # Save prototxt, caffemodel
        with open(prototxt_path, 'w') as f:    f.write(str(caffe_proto))
        with open(caffemodel_path, 'wb') as f: f.write(caffe_model.SerializeToString())

    def export(self, file_path: str, graph: BaseGraph, config_path: str=None, input_shapes: List[List[int]]=[[1,3,224,224]]):
        # dump config
        if config_path is not None:
            self.export_quantization_config(config_path=config_path, graph=graph)

        # dump model
        caffe_model, caffe_proto = self.prepare_model(graph, input_shapes)

        caffe_proto, str_buffer = str(caffe_proto), ''
        lines = caffe_proto.split('\n')

        for idx in range(len(lines)):
            line = lines[idx]
            # snpe do not want hole and ceil_mode
            if 'hole' in line or 'ceil_mode' in line: continue
            # snpe do not want quantize_param
            if 'quantize_param' in line:
                idx += 4
                continue
            str_buffer = (str_buffer + line) + '\n'
        caffe_proto = str_buffer
        self.dump_to_file(caffe_model = caffe_model,
            caffe_proto = caffe_proto, file_path = file_path)


class SNPECaffeExporter(CaffeExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        var_quant_info_recorder = {}
        for operation in graph.operations.values():
            if not isinstance(operation, QuantableOperation): continue
            for config, var in operation.config_with_variable:
                if not QuantizationStates.can_export(config.state):
                    raise PermissionError(
                        'Can not export quant config cause not all quantization configurations '
                        'have been correctly initialized(or some of them has been deactivated). '
                        f'Operation {operation.name} has an invalid quantization config({config.state}) '
                        f'at variable {var.name}.')

                # PATCH 2021.11.25
                # REMOVE BIAS FROM CONFIGURATION
                if config.num_of_bits > 8:
                        continue

                if config.state in {
                    QuantizationStates.FP32,
                    QuantizationStates.SOI
                }: continue
                # Simply override recorder is acceptable here,
                # we do not support mix precision quantization for CUDA backend now.
                # All configurations for this variable should keep identical towards each other.
                if config.state == QuantizationStates.PASSIVE and var.name in var_quant_info_recorder: continue
                var_quant_info_recorder[var.name] = config

        # ready to render config to json.
        for var in var_quant_info_recorder:
            config = var_quant_info_recorder[var]
            assert isinstance(config, TensorQuantizationConfig)
            tensorwise = config.policy.has_property(QuantizationProperty.PER_TENSOR)
            var_quant_info_recorder[var] = {
                'tensor_min' : convert_value(config.scale * (config.quant_min - config.offset), tensorwise, DataType.FP32),
                'tensor_max' : convert_value(config.scale * (config.quant_max - config.offset), tensorwise, DataType.FP32)
            }

        with open(file=config_path, mode='w') as file:
            json.dump(var_quant_info_recorder, file, indent=4)

    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, input_shapes: List[List[int]] = [[1, 3, 224, 224]]):
        # dump config
        if config_path is not None:
            self.export_quantization_config(config_path=config_path, graph=graph)

        caffe_model, caffe_proto = self.prepare_model(graph, input_shapes)

        # can not edit structure of protobuf with python, have to edit it as pure string.
        caffe_proto, str_buffer = str(caffe_proto), ''
        lines = caffe_proto.split('\n')
        for idx in range(len(lines)):
            line = lines[idx]
            # snpe do not want hole and ceil_mode
            if 'hole' in line or 'ceil_mode' in line: continue
            # snpe do not want quantize_param
            if 'quantize_param' in line:
                idx += 4
                continue
            str_buffer = (str_buffer + line) + '\n'
        caffe_proto = str_buffer

        # dump model
        self.dump_to_file(
            caffe_model = caffe_model,
            caffe_proto = caffe_proto,
            file_path = file_path)


class PPLDSPCaffeExporter(CaffeExporter):
    def export(self, file_path: str, graph: BaseGraph, config_path: str = None,
               input_shapes: List[List[int]] = [[1, 3, 224, 224]], write_weight=False):
        # PPL3 DSP do not need a json config file, all quantization configuration will be merged into protobuf
        caffe_model, caffe_proto = self.prepare_model(graph, input_shapes)
        for idx in range(len(caffe_proto.layer)):
            layer = caffe_proto.layer[idx]
            layer_name = layer.name

            assert isinstance(layer, ppl_caffe_pb2.LayerParameter)
            assert isinstance(layer_name, str)
            # layer is a caffe data structure, corresponding to operation in ppq.
            # following code write ppq quantization configuration to caffe layer.

            # step - 1, find corresponding op
            if layer_name not in graph.operations:
                # PATCH FOR Slice
                if layer.type == 'Slice':
                    # slice0 --> (ppq parse, export, combine) --> slice0_0_0, everything else
                    # is the same with original model
                    # layer_name is slice0_0_0, obtain original_name slice0
                    original_name = ''.join(layer_name.split('_')[:-2])
                    for bottom in layer.bottom:
                        var = graph.variables.get(bottom, None)
                        if var is not None and isinstance(var, QuantableVariable) and not var.is_parameter:
                            cfg = None
                            for dest_op, dest_cfg in zip(var.dest_ops, var.dest_op_configs):
                                # dest_op.name is slice0_0
                                if ''.join(dest_op.name.split('_')[:-1]) == original_name:
                                    cfg = dest_cfg
                                    break
                            assert cfg is not None
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            layer.quantize_param.add(type='bottom', range_min=qt_min, range_max=qt_max)

                    for top in layer.top:
                        var = graph.variables.get(top, None)
                        if var is not None and isinstance(var, QuantableVariable) and not var.is_parameter:
                            cfg = var.source_op_config
                            assert cfg is not None
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            layer.quantize_param.add(type='top', range_min=qt_min, range_max=qt_max)
                    continue
                else:
                    raise KeyError(f'Can not find operation {layer_name} with current graph.')

            op = graph.operations[layer_name]
            if not isinstance(op, QuantableOperation): continue

            # step - 2, dump parameter quantization config
            if layer.type in {'Convolution', 'Deconvolution'}:
                for cfg, var in op.config_with_variable:
                    if not var.is_parameter: continue
                    if cfg.num_of_bits > 8: continue # skip bias

                    if cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), False, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), False, DataType.FP32)
                        for _min, _max in zip(qt_min, qt_max):
                            p = layer.convolution_param.perchannel_quantize_param.add()
                            p.type = 'filter'
                            p.range_min = _min
                            p.range_max = _max

                    if cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                        p = layer.convolution_param.quantize_param
                        p.type = 'filter'
                        p.range_min = qt_min
                        p.range_max = qt_max

            if layer.type == 'InnerProduct':
                for cfg, var in op.config_with_variable:
                    if not var.is_parameter: continue
                    if cfg.num_of_bits > 8: continue # skip bias

                    if cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), False, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), False, DataType.FP32)
                        for _min, _max in zip(qt_min, qt_max):
                            p = layer.inner_product_param.perchannel_quantize_param.add()
                            p.type = 'filter'
                            p.range_min = _min
                            p.range_max = _max

                    if cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                        p = layer.inner_product_param.quantize_param
                        p.type = 'filter'
                        p.range_min = qt_min
                        p.range_max = qt_max

            if layer.type == 'PReLU':
                for cfg, var in op.config_with_variable:
                    if not var.is_parameter: continue

                    if cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), False, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), False, DataType.FP32)
                        for _min, _max in zip(qt_min, qt_max):
                            p = layer.prelu_param.perchannel_quantize_param.add()
                            p.type = 'slope'
                            p.range_min = _min
                            p.range_max = _max

                    if cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                        p = layer.prelu_param.quantize_param
                        p.type = 'slope'
                        p.range_min = qt_min
                        p.range_max = qt_max

            # step - 3 dump input config
            # only ops related with graph inputs record bottom param
            # usually first conv op
            record_bottom = False
            for input_var in op.inputs:
                if input_var.name in graph.inputs:
                    record_bottom = True
                    break
            if record_bottom:
                for bottom in layer.bottom:
                    for cfg, var, in zip(op.config.input_quantization_config, op.inputs):
                        if var.name == str(bottom):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            if var.name in graph.inputs:
                                range_min = convert_value(cfg.detail.get('range_min', -1), True, DataType.FP32)
                                # no negative values in input
                                if range_min >= 0.0:
                                    qt_min = 0.0
                            layer.quantize_param.add(type='bottom', range_min=qt_min, range_max=qt_max)

            # step - 4 dump output config
            for top in layer.top:
                for cfg, var, in zip(op.config.output_quantization_config, op.outputs):
                    assert cfg.policy.has_property(QuantizationProperty.PER_TENSOR)

                    if var.name == str(top):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                        layer.quantize_param.add(type='top', range_min=qt_min, range_max=qt_max)

        # dump model
        self.dump_to_file(
            caffe_model = caffe_model,
            caffe_proto = caffe_proto,
            file_path = file_path)


class PPLDSPTICaffeExporter(CaffeExporter):
    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, \
                input_shapes: List[List[int]] = [[1, 3, 224, 224]], write_weight=False):
        # PPL3 DSP TI do not need a json config file, all quantization configuration will be merged into protobuf
        # DSP TI differs from DSP in the perchannel quantization param for computing ops
        caffe_model, caffe_proto = self.prepare_model(graph, input_shapes)
        for idx in range(len(caffe_proto.layer)):
            layer = caffe_proto.layer[idx]
            layer_name = layer.name

            assert isinstance(layer, ppl_caffe_pb2.LayerParameter)
            assert isinstance(layer_name, str)
            # layer is a caffe data structure, corresponding to operation in ppq.
            # following code write ppq quantization configuration to caffe layer.

             # step - 1, find corresponding op
            if layer_name not in graph.operations:
                # PATCH FOR Slice
                if layer.type == 'Slice':
                    # slice0 --> (ppq parse, export, combine) --> slice0_0_0, everything else
                    # is the same with original model
                    # layer_name is slice0_0_0, obtain original_name slice0
                    original_name = ''.join(layer_name.split('_')[:-2])
                    for bottom in layer.bottom:
                        var = graph.variables.get(bottom, None)
                        if var is not None and isinstance(var, QuantableVariable) and not var.is_parameter:
                            cfg = None
                            for dest_op, dest_cfg in zip(var.dest_ops, var.dest_op_configs):
                                # dest_op.name is slice0_0
                                if ''.join(dest_op.name.split('_')[:-1]) == original_name:
                                    cfg = dest_cfg
                                    break
                            assert cfg is not None
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            layer.quantize_param.add(type='bottom', range_min=qt_min, range_max=qt_max)

                    for top in layer.top:
                        var = graph.variables.get(top, None)
                        if var is not None and isinstance(var, QuantableVariable) and not var.is_parameter:
                            cfg = var.source_op_config
                            assert cfg is not None
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            layer.quantize_param.add(type='top', range_min=qt_min, range_max=qt_max)
                    continue
                else:
                    raise KeyError(f'Can not find operation {layer_name} with current graph.')

            op = graph.operations[layer_name]
            if not isinstance(op, QuantableOperation): continue

            # don't dump by default
            if write_weight:
                if layer.type in {'Convolution', 'Deconvolution'}:
                    for cfg, var in op.config_with_variable:
                        if not var.is_parameter: continue
                        if cfg.num_of_bits > 8: continue # skip bias

                        if cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), False, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), False, DataType.FP32)
                            for _min, _max in zip(qt_min, qt_max):
                                p = layer.convolution_param.perchannel_quantize_param.add()
                                p.type = 'filter'
                                p.range_min = _min
                                p.range_max = _max

                        if cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            p = layer.convolution_param.quantize_param
                            p.type = 'filter'
                            p.range_min = qt_min
                            p.range_max = qt_max

                if layer.type == 'InnerProduct':
                    for cfg, var in op.config_with_variable:
                        if not var.is_parameter: continue
                        if cfg.num_of_bits > 8: continue # skip bias

                        if cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), False, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), False, DataType.FP32)
                            for _min, _max in zip(qt_min, qt_max):
                                p = layer.inner_product_param.perchannel_quantize_param.add()
                                p.type = 'filter'
                                p.range_min = _min
                                p.range_max = _max

                        if cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            p = layer.inner_product_param.quantize_param
                            p.type = 'filter'
                            p.range_min = qt_min
                            p.range_max = qt_max

                if layer.type == 'PReLU':
                    for cfg, var in op.config_with_variable:
                        if not var.is_parameter: continue

                        if cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), False, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), False, DataType.FP32)
                            for _min, _max in zip(qt_min, qt_max):
                                p = layer.prelu_param.perchannel_quantize_param.add()
                                p.type = 'slope'
                                p.range_min = _min
                                p.range_max = _max

                        if cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            p = layer.prelu_param.quantize_param
                            p.type = 'slope'
                            p.range_min = qt_min
                            p.range_max = qt_max

            # step - 3 dump input config
            # only ops related with graph inputs record bottom param
            # usually first conv op
            record_bottom = False
            for input_var in op.inputs:
                if input_var.name in graph.inputs:
                    record_bottom = True
                    break
            if record_bottom:
                for bottom in layer.bottom:
                    for cfg, var, in zip(op.config.input_quantization_config, op.inputs):
                        if var.name == str(bottom):
                            qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                            qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                            if var.name in graph.inputs:
                                range_min = convert_value(cfg.detail.get('range_min', -1), True, DataType.FP32)
                                # no negative values in input
                                if range_min >= 0.0:
                                    qt_min = 0.0
                            layer.quantize_param.add(type='bottom', range_min=qt_min, range_max=qt_max)

            # step - 4 dump output config
            for top in layer.top:
                for cfg, var, in zip(op.config.output_quantization_config, op.outputs):
                    assert cfg.policy.has_property(QuantizationProperty.PER_TENSOR)

                    if var.name == str(top):
                        qt_min = convert_value(cfg.scale * (cfg.quant_min - cfg.offset), True, DataType.FP32)
                        qt_max = convert_value(cfg.scale * (cfg.quant_max - cfg.offset), True, DataType.FP32)
                        layer.quantize_param.add(type='top', range_min=qt_min, range_max=qt_max)

                    if op.is_computing_op:
                        for _min, _max in zip(cfg.detail['range_min'], cfg.detail['range_max']):
                            layer.quantize_param.add(type='topperchannel', range_min=_min, range_max=_max)

        # dump model
        self.dump_to_file(
            caffe_model = caffe_model,
            caffe_proto = caffe_proto,
            file_path = file_path)



"""这个脚本将教会你如何使用 PPQ 量化自定义算子"""

import torch
from ppq import *
from ppq.api import *
from ppq.quantization.quantizer import TensorRTQuantizer

B = 1
T = 64
MODEL_PATH = 'models\encoder_ln.onnx'

def generate_samples(num_of_samples: int = 32):
    """生成样本数据，把这个函数改成真实数据读入就可以完成量化了
    这个语音数据量很小 我建议你把整个数据集直接全部送上CUDA
    """
    sample = {
        'speech': torch.rand(size=[B, T, 80]).float().cuda(),
        'speech_lengths': torch.ones(size=[B]).int().cuda()}
    samples = [sample for _ in range(num_of_samples)]
    return samples
SAMPLES = generate_samples()

# 定义一个自己的量化器，定制量化行为，继承于 TensorRTQuantizer 量化器
class MyTensorRTQuantizer(TensorRTQuantizer):
    @ property
    def quant_operation_types(self) -> set:
        """覆盖 quant_operation_types  自定义需要量化的算子"""
        return {'LayerNormPlugin'}

    def init_quantize_config(
        self, operation: Operation) -> OperationQuantizationConfig:
        config = super().init_quantize_config(operation=operation)
        """针对 LayerNormPlugin 生成量化配置信息"""
        if operation.type == 'LayerNormPlugin': 
            wconfig = config.input_quantization_config[1] # weight config
            bconfig = config.input_quantization_config[2]
            
            wconfig.policy = QuantizationPolicy(    # weight 做 Per channel 量化
                QuantizationProperty.SYMMETRICAL +
                QuantizationProperty.LINEAR +
                QuantizationProperty.PER_TENSOR)
            bconfig.state = QuantizationStates.FP32 # bias 不量化
            '''
            # 将 weight config 升级为 ChannelwiseTensorQuantizationConfig
            config.input_quantization_config[1] = ( 
                ChannelwiseTensorQuantizationConfig.convert_from_tensor_config(
                    convert_from = wconfig,
                    offsets = None, scales = None, 
                    channel_axis = 0))
            '''
            config.input_quantization_config[1].observer_algorithm = 'Minmax'
        return config

def layernorm_forward(op: Operation, values: List[torch.Tensor], ctx = None, **kwargs):
    """自定义算子的前向传播函数"""
    return values[0]

register_operation_handler(
    handler=layernorm_forward, 
    operation_type='LayerNormPlugin', 
    platform=TargetPlatform.TRT_INT8)

register_network_quantizer(
    quantizer=MyTensorRTQuantizer,
    platform=TargetPlatform.TRT_INT8)

with ENABLE_CUDA_KERNEL():
    QS = QuantizationSettingFactory.trt_setting()
    ir = load_onnx_graph(onnx_import_file=MODEL_PATH)
    # 默认调度失效，直接手动调度所有 LayerNormPlugin 送上量化区
    for op in ir.operations.values():
        if op.type == 'LayerNormPlugin':
            QS.dispatching_table.append(operation=op.name, platform=TargetPlatform.TRT_INT8)
    
    qir = quantize_native_model(
        model=ir, calib_dataloader=SAMPLES, calib_steps=32, 
        input_shape=None, inputs=SAMPLES[0], 
        platform=TargetPlatform.TRT_INT8, setting=QS)

    graphwise_error_analyse(
        graph=qir, running_device='cuda', 
        dataloader=SAMPLES)

    export_ppq_graph(
        graph=qir, platform=TargetPlatform.ONNXRUNTIME, 
        graph_save_to='quantized.onnx')



from typing import Union

import torch
from ppq.core import (PASSIVE_OPERATIONS, OperationQuantizationConfig,
                      QuantizationPolicy, QuantizationProperty,
                      QuantizationStates, RoundingPolicy, TargetPlatform)
from ppq.IR import BaseGraph, Operation

from .base import BaseQuantizer


class OpenvinoQuantizer(BaseQuantizer):
    def __init__(
        self, graph: BaseGraph
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = - 128
        self._quant_max = + 127

    def init_quantize_config(
        self, operation: Operation) -> OperationQuantizationConfig:
        OQC = self.create_default_quant_config(
            policy=self.quantize_policy, rounding=self.rounding_policy,
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=0,
            quant_max=self._quant_max, quant_min=self._quant_min,
            observer_algorithm='percentile'
        )

        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul', 'PPQBiasFusedMatMul'}:
            # base_quant_config.output_quantization_config[0].state = QuantizationStates.FP32
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {'Conv', 'ConvTranspose'}:
                if operation.inputs[1].is_parameter:
                    conv_weight_config = OQC.input_quantization_config[1]
                    conv_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL +
                        QuantizationProperty.LINEAR +
                        QuantizationProperty.PER_CHANNEL
                    )
                    conv_weight_config.channel_axis = (1 if operation.type == 'ConvTranspose' else 0)
                    conv_weight_config.observer_algorithm = 'minmax'

            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type in {'Gemm', 'MatMul', 'PPQBiasFusedMatMul'}:
                if operation.inputs[1].is_parameter:
                    gemm_weight_config = OQC.input_quantization_config[1]
                    gemm_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL +
                        QuantizationProperty.LINEAR +
                        QuantizationProperty.PER_CHANNEL
                    )
                    gemm_weight_config.channel_axis = 0
                    gemm_weight_config.observer_algorithm = 'minmax'

            if operation.num_of_input > 2:
                bias_config = OQC.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type == 'LayerNormalization':
            # Layernorm - gamma and beta need to be FP32
            for TQC in OQC.input_quantization_config[1: ]:
                TQC.state = QuantizationStates.FP32

        if operation.type == 'Attention':
            # Attention - Only input and weight need to be quantized.
            for TQC in OQC.input_quantization_config[2: ]:
                TQC.state = QuantizationStates.FP32

        return OQC

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.OPENVINO_INT8

    @ property
    def quant_operation_types(self) -> set:
        return {
            'Conv', 'ConvTranspose', 'Gemm', 'Relu', 'PRelu',
            'Clip', 'Pad', 'Resize', 'MaxPool', 'AveragePool',
            'GlobalMaxPool', 'GlobalAveragePool', 'Softmax',
            'Mul', 'Add', 'Max', 'Sub', 'Div', 'Reshape',
            'LeakyRelu', 'Concat', 'Sigmoid', 'Interp',
            'ReduceMean', 'Transpose', 'Slice', 'Flatten',
            'HardSwish', 'HardSigmoid', 'MatMul',
            'Attention', 'LayerNormalization', 'Gelu',
            'PPQBiasFusedMatMul'
        }

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL +
            QuantizationProperty.LINEAR +
            QuantizationProperty.PER_TENSOR
        )

    @ property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @ property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Swish', 'Clip', 'SoftPlus', 'Sigmoid', 'Gelu'}


from typing import Union

import torch
from ppq.core import (PASSIVE_OPERATIONS, OperationQuantizationConfig,
                      QuantizationPolicy, QuantizationProperty,
                      QuantizationStates, RoundingPolicy, TargetPlatform)
from ppq.IR import BaseGraph, Operation
from .base import BaseQuantizer

from ppq.executor.op.torch.base import GET_ATTRIBUTE_FROM_OPERATION

def ASSERT_CONV_AND_DECONV(op: Operation):
    if op.type == "Conv":
        filter = op.inputs[1].value.shape
        assert len(filter) == 4, (
            f'Ascend Quantization needs the dimension of filter must be 4, while your filter is {len(filter)}')


    if op.type == "ConvTranspose":
        filter = op.inputs[1].value.shape
        dilation  = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='dilations', compulsive=True, default=1)
        groups    = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='group', compulsive=True, default=1)

        assert len(filter) == 4, (
            f'Ascend Quantization needs the dimension of filter must be 4, while your filter is {len(filter)}')
        assert dilation == [1, 1], (
            f'Ascend Quantization needs the dilation must be [1, 1], while your dilation is {dilation}')
        assert groups == 1, (
            f'Ascend Quantization needs the groups must be 1, while your groups is {groups}')



def ASSERT_GEMM(op: Operation):
    axis  = GET_ATTRIBUTE_FROM_OPERATION(op=op, attribute='axis', compulsive=True, default=1)
    assert axis == 1, (
            f'Ascend Quantization needs the axis must be 1, while your axis is {axis}')

    if 'transpose' in op.attributes:
        transpose = op.attributes['transpose']
        assert transpose == False, (
            f'Ascend Quantization needs the transpose must be False, while your transpose is {transpose}')



class AscendQuantizer(BaseQuantizer):
    def __init__(
        self,
        graph: BaseGraph
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = 0
        self._quant_max = int(pow(2, self._num_of_bits) - 1)


    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=0,
            quant_max= self._quant_max, quant_min= self._quant_min, observer_algorithm='percentile', 
            policy=self.quantize_policy, rounding=self.rounding_policy,
        )

        ASSERT_CONV_AND_DECONV(operation)

        if operation.type in {'Conv', 'Gemm', 'ConvTranspose'}:

            assert operation.num_of_input > 0, 'Seems you got a Computing layer with no parameters.'

            if operation.type in {'Conv', 'ConvTranspose'}:

                if operation.inputs[1].is_parameter:
                    conv_weight_config = base_quant_config.input_quantization_config[1]
                    conv_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL +
                        QuantizationProperty.LINEAR +
                        QuantizationProperty.PER_CHANNEL
                    )
                    conv_weight_config.channel_axis = (1 if operation.type == 'ConvTranspose' else 0)
                    conv_weight_config.observer_algorithm = 'minmax'
                    conv_weight_config.quant_max = 127
                    conv_weight_config.quant_min = -128

            elif operation.type == 'Gemm':
                
                ASSERT_GEMM(operation)

                if operation.inputs[1].is_parameter:
                    gemm_weight_config = base_quant_config.input_quantization_config[1]
                    gemm_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL +
                        QuantizationProperty.LINEAR +
                        QuantizationProperty.PER_TENSOR
                    )
                    
                    # gemm_weight_config.channel_axis = 0
                    gemm_weight_config.observer_algorithm = 'minmax'
                    gemm_weight_config.quant_max = 127
                    gemm_weight_config.quant_min = -128

            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.ASC_INT8

    @ property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @ property
    def quant_operation_types(self) -> set:
        return {'Conv', 'ConvTranspose', 'Gemm', 'AveragePool'}

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL +
            QuantizationProperty.LINEAR +
            QuantizationProperty.PER_TENSOR
        )

    @ property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO

    @ property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}

